import os
import gzip
import json

from io import BytesIO
from functools import total_ordering

import pickle

from six import string_types as basestring

import sqlalchemy
from sqlalchemy import (
    Column, Integer, String, ForeignKey,
    PickleType, Boolean, Table, ForeignKeyConstraint,
    BLOB)

from sqlalchemy.types import TypeDecorator
from sqlalchemy.orm import relationship, deferred
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.mutable import MutableDict

from ms_deisotope.data_source._compression import starts_with_gz_magic

try:
    from ms_deisotope.data_source._compression import starts_with_zstd_magic
except ImportError:
    starts_with_zstd_magic = None

try:
    from pyzstd import ZstdFile
except ImportError:
    ZstdFile = None


from glypy import Composition

from ..base import (
    Base)

from ..utils import get_or_create


def _ipython_key_completions_(self):
    return list(self.keys())


MutableDict._ipython_key_completions_ = _ipython_key_completions_


@total_ordering
class ApplicationVersion(Base):
    __tablename__ = "ApplicationVersion"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    major = Column(Integer)
    minor = Column(Integer)
    patch = Column(Integer)

    def __eq__(self, other):
        if other is None:
            return False
        try:
            if self.name != other.name:
                return False
        except AttributeError:
            pass

        return tuple(self) == tuple(other)

    def __lt__(self, other):
        if other is None:
            return False
        try:
            if self.name != other.name:
                return False
        except AttributeError:
            pass

        return tuple(self) < tuple(other)

    def __iter__(self):
        yield self.major
        yield self.minor
        yield self.patch


class ParameterStore(object):
    parameters = Column(MutableDict.as_mutable(PickleType), default=dict)


class Taxon(Base):
    __tablename__ = "Taxon"
    id = Column(Integer, primary_key=True)
    name = Column(String(128), index=True)

    @classmethod
    def get(cls, session, id, name=None):
        obj, made = get_or_create(session, cls, id=id, name=name)
        return obj

    def __repr__(self):
        return "<Taxon {} {}>".format(self.id, self.name)


class HasTaxonomy(object):

    @declared_attr
    def taxa(cls):
        taxon_association = Table(
            "%s_Taxa" % cls.__tablename__,
            cls.metadata,
            Column("taxon_id", Integer, ForeignKey(Taxon.id, ondelete="CASCADE"), primary_key=True),
            Column("entity_id", Integer, ForeignKey(
                "%s.id" % cls.__tablename__, ondelete="CASCADE"), primary_key=True))
        cls.TaxonomyAssociationTable = taxon_association
        return relationship(Taxon, secondary=taxon_association)

    @classmethod
    def with_taxa(cls, ids):
        try:
            iter(ids)
            return cls.taxa.any(Taxon.id.in_(tuple(ids)))
        except Exception:
            return cls.taxa.any(Taxon.id == ids)


class ReferenceDatabase(Base):
    __tablename__ = "ReferenceDatabase"
    id = Column(Integer, primary_key=True)
    name = Column(String(128))
    url = Column(String(128))

    @classmethod
    def get(cls, session, id=None, name=None, url=None):
        obj, made = get_or_create(session, cls, id=id, name=name, url=url)
        return obj

    def __repr__(self):
        return "<ReferenceDatabase {} {}>".format(self.id, self.name)


class ReferenceAccessionNumber(Base):
    __tablename__ = "ReferenceAccessionNumber"
    id = Column(String(64), primary_key=True)
    database_id = Column(Integer, ForeignKey(ReferenceDatabase.id), primary_key=True)
    database = relationship(ReferenceDatabase)

    @classmethod
    def get(cls, session, id, database_id):
        obj, made = get_or_create(session, cls, id=id, database_id=database_id)
        return obj

    def __repr__(self):
        return "<ReferenceAccessionNumber {} {}>".format(self.id, self.database.name)


class HasReferenceAccessionNumber(object):
    @declared_attr
    def references(cls):
        reference_number_association = Table(
            "%s_ReferenceAccessionNumber" % cls.__tablename__,
            cls.metadata,
            Column("accession_code", String(64), primary_key=True),
            Column("database_id", Integer, primary_key=True),
            Column(
                "entity_id", Integer, ForeignKey("%s.id" % cls.__tablename__, ondelete="CASCADE"), primary_key=True),
            ForeignKeyConstraint(
                ["accession_code", "database_id"],
                ["ReferenceAccessionNumber.id", "ReferenceAccessionNumber.database_id"]))
        cls.ReferenceAccessionAssocationTable = reference_number_association
        return relationship(ReferenceAccessionNumber, secondary=reference_number_association)


TemplateNumberStore = Table("TemplateNumberStore", Base.metadata, Column("value", Integer))


class FileBlob(Base):
    __tablename__ = 'FileBlob'

    id = Column(Integer, primary_key=True)
    name = Column(String(256), index=True)
    data = deferred(Column(BLOB))
    compressed = Column(Boolean)

    def __repr__(self):
        return "FileBlob({self.name})".format(self=self)

    @classmethod
    def from_file(cls, fp, compress=False):
        inst = cls(name=os.path.basename(fp.name), compressed=compress)
        data = fp.read()
        has_gz_magic = starts_with_gz_magic(data)
        if starts_with_zstd_magic is None:
            has_zstd_magic = False
        else:
            has_zstd_magic = starts_with_zstd_magic(data)
        if compress and not has_gz_magic and not has_zstd_magic:
            buff = BytesIO()
            with gzip.GzipFile(mode='wb', fileobj=buff) as compressor:
                compressor.write(data)
            buff.seek(0)
            data = buff.read()
            inst.name += '.gz'
        if not compress and has_gz_magic:
            inst.compressed = True
        if not compress and has_zstd_magic:
            inst.compressed = True
        inst.data = data
        return inst

    @classmethod
    def from_path(cls, path, compress=False):
        with open(path, 'rb') as fh:
            inst = cls.from_file(fh, compress)
        return inst

    def open(self, raw=False):
        data_buffer = BytesIO(self.data)
        if self.compressed and not raw:
            if starts_with_gz_magic(self.data[:6]):
                return gzip.GzipFile(fileobj=data_buffer)
            elif (starts_with_zstd_magic is not None and ZstdFile is not None) and starts_with_zstd_magic(self.data[:6]):
                return ZstdFile(data_buffer)
            else:
                raise ValueError("Could not infer decompressor for %r" % (self.data[:6], ))
        return data_buffer


class HasFiles(object):
    @declared_attr
    def files(cls):
        file_association = Table(
            "%s_Files" % cls.__tablename__,
            cls.metadata,
            Column("file_id", Integer, ForeignKey(FileBlob.id, ondelete="CASCADE"), primary_key=True),
            Column("entity_id", Integer, ForeignKey(
                "%s.id" % cls.__tablename__, ondelete="CASCADE"), primary_key=True))
        cls.FileBlobAssociationTable = file_association
        return relationship(FileBlob, secondary=file_association)

    def add_file(self, file_obj, compress=False):
        if isinstance(file_obj, basestring):
            f = FileBlob.from_path(file_obj, compress)
        else:
            f = FileBlob.from_file(file_obj, compress)
        self.files.append(f)


class JSONType(TypeDecorator):

    impl = sqlalchemy.Text()

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = "J" + json.dumps(value)

        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            if isinstance(value, bytes):
                # NOTE: This should only happen when loading a database originally written in Py2?
                if value.startswith(b"J"):
                    value = json.loads(value[1:].decode('utf8'))
                else:
                    value = pickle.loads(value)
            else:
                value = json.loads(value[1:])
        return value


class HasChemicalComposition(object):
    _total_composition = None

    def total_composition(self):
        if self._total_composition is None:
            self._total_composition = Composition(self.formula)
        return self._total_composition



import os
import sys
import time
import logging

import click

from glycresoft.cli.base import cli
from glycresoft.cli.validators import get_by_name_or_id, DatabaseConnectionParam

from glycresoft import serialize
from glycresoft.serialize import (
    DatabaseBoundOperation, GlycanHypothesis, GlycopeptideHypothesis,
    Analysis,
    AnalysisTypeEnum,
    GlycanCompositionChromatogram,
    Protein,
    Glycopeptide,
    IdentifiedGlycopeptide,
    GlycopeptideSpectrumMatch,
    AnalysisDeserializer,
    GlycanComposition,
    GlycanCombination,
    GlycanCombinationGlycanComposition,
    UnidentifiedChromatogram)

from glycresoft.output import (
    GlycanHypothesisCSVSerializer, ImportableGlycanHypothesisCSVSerializer,
    GlycopeptideHypothesisCSVSerializer, GlycanLCMSAnalysisCSVSerializer,
    GlycopeptideLCMSMSAnalysisCSVSerializer,
    GlycopeptideSpectrumMatchAnalysisCSVSerializer,
    MultiScoreGlycopeptideLCMSMSAnalysisCSVSerializer,
    MultiScoreGlycopeptideSpectrumMatchAnalysisCSVSerializer,
    MzIdentMLSerializer,
    GlycanChromatogramReportCreator,
    GlycopeptideDatabaseSearchReportCreator,
    TrainingMGFExporter,
    SpectrumAnnotatorExport,
    CSVSpectrumAnnotatorExport)

from glycresoft.output.csv_format import csv_stream

from glycresoft.cli.utils import ctxstream


status_logger = logging.getLogger('glycresoft.status')


def database_connection_arg(fn):
    arg = click.argument(
        "database-connection",
        type=DatabaseConnectionParam(exists=True),
        doc_help=(
            "A connection URI for a database, or a path on the file system"))
    return arg(fn)


def analysis_identifier_arg(analysis_type):
    def wrapper(fn):
        arg = click.argument("analysis-identifier", doc_help=(
            "The ID number or name of the %s analysis to use" % (analysis_type,)))
        return arg(fn)
    return wrapper


def hypothesis_identifier_arg(hypothesis_type):
    def wrapper(fn):
        arg = click.argument("hypothesis-identifier", doc_help=(
            "The ID number or name of the %s hypothesis to use" % (hypothesis_type,)))
        return arg(fn)
    return wrapper


@cli.group(short_help='Write Data Collections To Text Files')
def export():
    pass


@export.command("glycan-hypothesis",
                short_help="Export theoretical glycan composition database to CSV")
@database_connection_arg
@hypothesis_identifier_arg("glycan")
@click.option("-o", "--output-path", type=click.Path(), default=None, help='Path to write to instead of stdout')
@click.option("-i", "--importable", is_flag=True,
              help="Make the file importable for later re-use, with less information.")
def glycan_hypothesis(database_connection, hypothesis_identifier, output_path=None, importable=False):
    """Write each theoretical glycan composition in CSV format"""
    database_connection = DatabaseBoundOperation(database_connection)
    hypothesis = get_by_name_or_id(database_connection, GlycanHypothesis, hypothesis_identifier)
    if importable:
        task_type = ImportableGlycanHypothesisCSVSerializer
    else:
        task_type = GlycanHypothesisCSVSerializer
    if output_path is None or output_path == '-':
        output_stream = ctxstream(sys.stdout)
    else:
        output_stream = open(output_path, 'wb')
    with output_stream:
        job = task_type(output_stream, hypothesis.glycans)
        job.run()


@export.command("glycopeptide-hypothesis",
                short_help='Export theoretical glycopeptide database to CSV')
@database_connection_arg
@hypothesis_identifier_arg("glycopeptide")
@click.option("-o", "--output-path", type=click.Path(), default=None, help='Path to write to instead of stdout')
def glycopeptide_hypothesis(database_connection, hypothesis_identifier, output_path, multifasta=False):
    """Write each theoretical glycopeptide in CSV format"""
    database_connection = DatabaseBoundOperation(database_connection)
    session = database_connection.session()  # pylint: disable=not-callable
    hypothesis = get_by_name_or_id(session, GlycopeptideHypothesis, hypothesis_identifier)

    def generate():
        interval = 100000
        i = 0
        while True:
            session.expire_all()
            chunk = hypothesis.glycopeptides.slice(i, i + interval).all()
            if len(chunk) == 0:
                break
            for glycopeptide in chunk:
                yield glycopeptide
            i += interval
    if output_path is None or output_path == '-':
        output_stream = ctxstream(sys.stdout)
    else:
        output_stream = open(output_path, 'wb')
    with output_stream:
        job = GlycopeptideHypothesisCSVSerializer(output_stream, generate())
        job.run()


@export.command("glycan-identification",
                short_help="Exports assigned LC-MS features of Glycan Compositions to CSV or HTML")
@database_connection_arg
@analysis_identifier_arg("glycan")
@click.option("-o", "--output-path", type=click.Path(), default=None, help='Path to write to instead of stdout')
@click.option("-r", "--report", is_flag=True, help="Export an HTML report instead of a CSV")
@click.option("-t", "--threshold", type=float, default=0, help="Minimum score threshold to include")
def glycan_composition_identification(database_connection, analysis_identifier, output_path=None,
                                      threshold=0, report=False):
    """Write each glycan chromatogram in CSV format"""
    database_connection = DatabaseBoundOperation(database_connection)
    session = database_connection.session()  # pylint: disable=not-callable
    analysis = get_by_name_or_id(session, Analysis, analysis_identifier)
    if not analysis.analysis_type == AnalysisTypeEnum.glycan_lc_ms:
        click.secho("Analysis %r is of type %r." % (
            str(analysis.name), str(analysis.analysis_type)), fg='red', err=True)
        raise click.Abort()
    analysis_id = analysis.id
    if output_path is None or output_path == '-':
        output_stream = ctxstream(sys.stdout)
    else:
        output_stream = open(output_path, 'wb')

    if report:
        with output_stream:
            job = GlycanChromatogramReportCreator(
                database_connection._original_connection,
                analysis_id, output_stream, threshold=threshold)
            job.run()
    else:
        def generate():
            i = 0
            interval = 100
            query = session.query(GlycanCompositionChromatogram).filter(
                GlycanCompositionChromatogram.analysis_id == analysis_id,
                GlycanCompositionChromatogram.score > threshold)

            while True:
                session.expire_all()
                chunk = query.slice(i, i + interval).all()
                if len(chunk) == 0:
                    break
                for gcs in chunk:
                    yield gcs.convert()
                i += interval

            i = 0
            query = session.query(UnidentifiedChromatogram).filter(
                UnidentifiedChromatogram.analysis_id == analysis_id,
                UnidentifiedChromatogram.score > threshold)

            while True:
                session.expire_all()
                chunk = query.slice(i, i + interval).all()
                if len(chunk) == 0:
                    break
                for gcs in chunk:
                    yield gcs.convert()
                i += interval

        with output_stream:
            job = GlycanLCMSAnalysisCSVSerializer(output_stream, generate())
            job.run()


@export.command("glycopeptide-identification",
                short_help="Exports assigned LC-MS/MS features of Glycopeptides to CSV or HTML")
@database_connection_arg
@analysis_identifier_arg("glycopeptide")
@click.option("-o", "--output-path", type=click.Path(), default=None, help='Path to write to instead of stdout')
@click.option("-r", "--report", is_flag=True, help="Export an HTML report instead of a CSV")
@click.option("-m", "--mzml-path", type=click.Path(exists=True), default=None, help=(
    "Path to read processed spectra from instead of the path embedded in the analysis metadata"))
@click.option("-t", "--threshold", type=float, default=0, help='The minimum MS2 score threshold (larger is better)')
@click.option("-f", "--fdr-threshold", type=float, default=1, help='The maximum FDR threshold (smaller is better)')
def glycopeptide_identification(database_connection, analysis_identifier, output_path=None,
                                report=False, mzml_path=None, threshold=0, fdr_threshold=1):
    """Write each distinct identified glycopeptide in CSV format"""
    # logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)
    database_connection = DatabaseBoundOperation(database_connection)
    session = database_connection.session()  # pylint: disable=not-callable
    analysis = get_by_name_or_id(session, Analysis, analysis_identifier)
    if not analysis.analysis_type == AnalysisTypeEnum.glycopeptide_lc_msms:
        click.secho("Analysis %r is of type %r." % (
            str(analysis.name), str(analysis.analysis_type)), fg='red', err=True)
        raise click.Abort()
    analysis_id = analysis.id
    is_multiscore = analysis.is_multiscore
    if output_path is None or output_path == '-':
        output_stream = ctxstream(click.get_binary_stream('stdout'))
    else:
        output_stream = open(output_path, 'wb')
    if report:
        with output_stream:
            if mzml_path is None:
                mzml_path = analysis.parameters['sample_path']
                if not os.path.exists(mzml_path):
                    raise click.ClickException(
                        ("Sample path {} not found. Pass the path to"
                         " this file as `-m/--mzml-path` for this command.").format(
                             mzml_path))
            job = GlycopeptideDatabaseSearchReportCreator(
                database_connection._original_connection, analysis_id,
                stream=output_stream,
                threshold=threshold,
                fdr_threshold=fdr_threshold,
                mzml_path=mzml_path)
            job.run()
    else:
        query = session.query(Protein.id, Protein.name).join(Protein.glycopeptides).join(
            IdentifiedGlycopeptide).filter(
                IdentifiedGlycopeptide.analysis_id == analysis.id)
        protein_index = dict(query)

        if is_multiscore:
            job_type = MultiScoreGlycopeptideLCMSMSAnalysisCSVSerializer
        else:
            job_type = GlycopeptideLCMSMSAnalysisCSVSerializer

        def generate():
            i = 0
            interval = 100
            query = session.query(IdentifiedGlycopeptide).filter(
                IdentifiedGlycopeptide.analysis_id == analysis_id)
            if threshold != 0:
                query = query.filter(IdentifiedGlycopeptide.ms2_score >= threshold)
            if fdr_threshold < 1:
                query = query.filter(IdentifiedGlycopeptide.q_value <= fdr_threshold)
            while True:
                session.expunge_all()
                chunk = query.slice(i, i + interval).all()
                if len(chunk) == 0:
                    break
                # chunk = IdentifiedGlycopeptide.bulk_convert(chunk)
                for glycopeptide in chunk:
                    yield glycopeptide
                i += interval
        with output_stream:
            job = job_type(output_stream, generate(), protein_index, analysis)
            job.run()


@export.command("glycopeptide-spectrum-matches",
                short_help="Exports individual MS/MS assignments of Glycopeptides to CSV")
@database_connection_arg
@analysis_identifier_arg("glycopeptide")
@click.option("-o", "--output-path", type=click.Path(), default=None, help='Path to write to instead of stdout')
@click.option("-t", "--threshold", type=float, default=0, help='The minimum MS2 score threshold (larger is better)')
@click.option("-f", "--fdr-threshold", type=float, default=1, help='The maximum FDR threshold (smaller is better)')
def glycopeptide_spectrum_matches(database_connection, analysis_identifier, output_path=None, threshold=0,
                                  fdr_threshold=1.0):
    """Write each matched glycopeptide spectrum in CSV format"""
    database_connection = DatabaseBoundOperation(database_connection)
    session = database_connection.session()  # pylint: disable=not-callable
    analysis = get_by_name_or_id(session, Analysis, analysis_identifier)
    if not analysis.analysis_type == AnalysisTypeEnum.glycopeptide_lc_msms:
        click.secho("Analysis %r is of type %r." % (
            str(analysis.name), str(analysis.analysis_type)), fg='red', err=True)
        raise click.Abort()
    analysis_id = analysis.id
    is_multiscore = analysis.is_multiscore
    start = time.monotonic()
    protein_query = session.query(Protein.id, Protein.name).join(Protein.glycopeptides).join(
        GlycopeptideSpectrumMatch).filter(
            GlycopeptideSpectrumMatch.analysis_id == analysis.id)
    status_logger.info("Loading protein index")
    protein_index = dict(protein_query.all())

    def generate():
        i = 0
        interval = 100000
        gpsm_query = session.query(GlycopeptideSpectrumMatch.id).filter(
            GlycopeptideSpectrumMatch.analysis_id == analysis_id).order_by(
                GlycopeptideSpectrumMatch.scan_id)
        if threshold != 0:
            gpsm_query = gpsm_query.filter(GlycopeptideSpectrumMatch.score >= threshold)
        if fdr_threshold < 1:
            gpsm_query = gpsm_query.filter(GlycopeptideSpectrumMatch.q_value <= fdr_threshold)

        mass_shift_cache = {
            m.id: m.convert()
            for m in session.query(serialize.CompoundMassShift).all()
        }
        scan_cache = {}
        structure_cache = {}
        peptide_relation_cache = {}

        status_logger.info("Reading spectrum matches")
        j = 0
        while True:
            session.expire_all()
            chunk = gpsm_query.slice(i, i + interval).all()
            chunk = [x[0] for x in chunk]
            chunk = GlycopeptideSpectrumMatch.bulk_load(
                session,
                chunk,
                mass_shift_cache=mass_shift_cache,
                scan_cache=scan_cache,
                structure_cache=structure_cache,
                peptide_relation_cache=peptide_relation_cache
            )
            if not chunk:
                break
            chunk = [x[1] for x in sorted(chunk.items())]
            for glycopeptide in chunk:
                j += 1
                yield glycopeptide
            i += interval

    if is_multiscore:
        job_type = MultiScoreGlycopeptideSpectrumMatchAnalysisCSVSerializer
    else:
        job_type = GlycopeptideSpectrumMatchAnalysisCSVSerializer

    if output_path is None or output_path == '-':
        output_stream = ctxstream(click.get_binary_stream('stdout'))
    else:
        output_stream = open(output_path, 'wb')

    with output_stream:
        job = job_type(output_stream, generate(), protein_index, analysis)
        job.run()

    end = time.monotonic()
    elapsed = end - start
    status_logger.info(f"{elapsed:0.2f} seconds elapsed")


@export.command("mzid", short_help="Export a Glycopeptide Analysis as MzIdentML")
@database_connection_arg
@analysis_identifier_arg("glycopeptide")
@click.argument("output-path")
@click.option("--embed-protein-sequences/--exclude-protein-sequences", is_flag=True, default=True,
              help="Include protein sequences in the output file")
@click.option("-m", '--mzml-path', type=click.Path(exists=True), default=None,
              help="Alternative path to find the source mzML file")
def glycopeptide_mzidentml(database_connection, analysis_identifier, output_path=None,
                           mzml_path=None, embed_protein_sequences=True):
    """
    Write identified glycopeptides as mzIdentML file, and associated MSn spectra
    to a paired mzML file if the matched data are available. If an mzML file is written
    it will also contain the extracted ion chromatograms for each glycopeptide with an
    extracted elution profile.
    """
    database_connection = DatabaseBoundOperation(database_connection)
    session = database_connection.session()  # pylint: disable=not-callable
    analysis = get_by_name_or_id(session, Analysis, analysis_identifier)
    if not analysis.analysis_type == AnalysisTypeEnum.glycopeptide_lc_msms:
        click.secho("Analysis %r is of type %r." % (
            str(analysis.name), str(analysis.analysis_type)), fg='red', err=True)
        raise click.Abort()
    loader = AnalysisDeserializer(
        database_connection._original_connection, analysis_id=analysis.id)
    click.echo("Loading Identifications")
    # glycopeptides = loader.load_identified_glycopeptides()
    glycopeptides = loader.query(IdentifiedGlycopeptide).filter(
        IdentifiedGlycopeptide.analysis_id == analysis_identifier).all()
    with open(output_path, 'wb') as outfile:
        writer = MzIdentMLSerializer(
            outfile, glycopeptides, analysis, loader,
            source_mzml_path=mzml_path,
            embed_protein_sequences=embed_protein_sequences)
        writer.run()


@export.command("glycopeptide-training-mgf", short_help=("Export identified MS2 spectra as MGF with"
                                                         " matched glycopeptide as metadata"))
@database_connection_arg
@analysis_identifier_arg("glycopeptide")
@click.option("-o", "--output-path", type=click.Path(), default=None, help='Path to write to instead of stdout')
@click.option("-m", '--mzml-path', type=click.Path(exists=True), default=None,
              help="Alternative path to find the source mzML file")
@click.option("-t", "--threshold", type=float, default=None, help='Minimum MS2 score to export')
def glycopeptide_training_mgf(database_connection, analysis_identifier, output_path=None,
                              mzml_path=None, threshold=None):
    """
    Export identified glycopeptide spectrum matches for model training as
    an annotated MGF file.
    """
    database_connection = DatabaseBoundOperation(database_connection)
    session = database_connection.session()  # pylint: disable=not-callable
    analysis = get_by_name_or_id(session, Analysis, analysis_identifier)
    if not analysis.analysis_type == AnalysisTypeEnum.glycopeptide_lc_msms:
        click.secho("Analysis %r is of type %r." % (
            str(analysis.name), str(analysis.analysis_type)), fg='red', err=True)
        raise click.Abort()
    if output_path is None or output_path == '-':
        output_stream = ctxstream(click.get_binary_stream('stdout'))
    else:
        output_stream = open(output_path, 'wb')
    with output_stream:
        TrainingMGFExporter.from_analysis(
            database_connection, analysis.id, output_stream, mzml_path, threshold).run()


@export.command("identified-glycans-from-glycopeptides", short_help="Export glycans on identified glycopeptides")
@database_connection_arg
@analysis_identifier_arg("glycopeptide")
@click.option("-o", "--output-path", type=click.Path(), default=None, help='Path to write to instead of stdout')
def export_identified_glycans_from_glycopeptides(database_connection, analysis_identifier, output_path):
    """
    Export the glycans attached to any identified glycopeptides.

    The files are written in an importable format to be used to create a new glycan hypothesis.
    """
    database_connection = DatabaseBoundOperation(database_connection)
    session = database_connection.session()  # pylint: disable=not-callable
    analysis = get_by_name_or_id(session, Analysis, analysis_identifier)
    if not analysis.analysis_type == AnalysisTypeEnum.glycopeptide_lc_msms:
        click.secho("Analysis %r is of type %r." % (
            str(analysis.name), str(analysis.analysis_type)), fg='red', err=True)
        raise click.Abort()
    glycans = session.query(GlycanComposition).join(
        GlycanCombinationGlycanComposition).join(GlycanCombination).join(
            Glycopeptide,
            Glycopeptide.glycan_combination_id == GlycanCombination.id).join(
                IdentifiedGlycopeptide,
                IdentifiedGlycopeptide.structure_id == Glycopeptide.id).filter(
                    IdentifiedGlycopeptide.analysis_id == analysis.id).all()
    if output_path is None:
        output_stream = ctxstream(click.get_binary_stream('stdout'))
    else:
        output_stream = open(output_path, 'wb')
    with output_stream:
        job = ImportableGlycanHypothesisCSVSerializer(output_stream, glycans)
        job.run()


@export.command("annotate-matched-spectra",
                short_help="Exports graphical renderings of individual MS/MS assignments of glycopeptides to PDF")
@database_connection_arg
@analysis_identifier_arg("glycopeptide")
@click.option("-o", "--output-path", type=click.Path(), default=None, help='Path to write to instead of stdout')
@click.option("-m", '--mzml-path', type=click.Path(exists=True), default=None,
              help="Alternative path to find the source mzML file")
def annotate_matched_spectra(database_connection, analysis_identifier, output_path, mzml_path=None):
    """
    Export graphical renderings of individual MS/MS assignments of glycopeptides to PDF.

    The output file will contain one spectrum per PDF page.
    """
    database_connection = DatabaseBoundOperation(database_connection)
    session = database_connection.session()  # pylint: disable=not-callable
    analysis = get_by_name_or_id(session, Analysis, analysis_identifier)
    if not analysis.analysis_type == AnalysisTypeEnum.glycopeptide_lc_msms:
        click.secho("Analysis %r is of type %r." % (
            str(analysis.name), str(analysis.analysis_type)), fg='red', err=True)
        raise click.Abort()
    if output_path is None:
        output_path = os.path.dirname(database_connection._original_connection)

    task = SpectrumAnnotatorExport(
        database_connection._original_connection, analysis.id, output_path,
        mzml_path)
    task.display_header()
    task.start()


@export.command("write-csv-spectrum-library", short_help="Exports individual MS/MS assignments of glycopeptides to CSV")
@database_connection_arg
@analysis_identifier_arg("glycopeptide")
@click.option("-o", "--output-path", type=click.Path(), default=None, help='Path to write to instead of stdout')
@click.option("-m", '--mzml-path', type=click.Path(exists=True), default=None,
              help="Alternative path to find the source mzML file")
@click.option("-t", "--fdr-threshold", type=float, default=0.05)
def write_spectrum_library(database_connection, analysis_identifier, output_path, mzml_path=None, fdr_threshold=0.05):
    """
    Export individual MS/MS assignments of glycopeptides to CSV.

    Each row corresponds to a single product ion match, partitioned by
    scan ID. Unmatched peaks are ignored.
    """
    database_connection = DatabaseBoundOperation(database_connection)
    session = database_connection.session()  # pylint: disable=not-callable
    analysis = get_by_name_or_id(session, Analysis, analysis_identifier)
    if not analysis.analysis_type == AnalysisTypeEnum.glycopeptide_lc_msms:
        click.secho("Analysis %r is of type %r." % (
            str(analysis.name), str(analysis.analysis_type)), fg='red', err=True)
        raise click.Abort()
    if output_path is None:
        output_stream = ctxstream(click.get_binary_stream('stdout'))
    else:
        output_stream = open(output_path, 'wb')

    with output_stream:
        task = CSVSpectrumAnnotatorExport(
            database_connection._original_connection, analysis.id, output_stream,
            mzml_path, fdr_threshold)
        task.run()


@export.command("glycopeptide-chromatogram-records")
@database_connection_arg
@analysis_identifier_arg("glycopeptide")
@click.option("-o", "--output-path", type=click.Path(), default=None, help='Path to write to instead of stdout')
@click.option('-r', '--apex-time-range', type=(float, float), default=(0, float('inf')),
              help='The range of apex times to include')
@click.option("-t", "--fdr-threshold", type=float, default=0.05)
def glycopeptide_chromatogram_records(
        database_connection, analysis_identifier, output_path, apex_time_range=None, fdr_threshold=0.05):
    """
    Export a simplified table of glycopeptide chromatographic features, formatted for retention time
    modeling.
    """
    if apex_time_range is None:
        apex_time_range = (0, float('inf'))
    database_connection = DatabaseBoundOperation(database_connection)
    session = database_connection.session()  # pylint: disable=not-callable
    analysis = get_by_name_or_id(session, Analysis, analysis_identifier)
    if not analysis.analysis_type == AnalysisTypeEnum.glycopeptide_lc_msms:
        click.secho("Analysis %r is of type %r." % (
            str(analysis.name), str(analysis.analysis_type)), fg='red', err=True)
        raise click.Abort()
    if output_path is None:
        fh = click.get_binary_stream('stdout')
    else:
        fh = open(output_path, 'wb')
    idgps = session.query(
        IdentifiedGlycopeptide).filter(
            IdentifiedGlycopeptide.analysis_id == analysis.id).all()
    n = len(idgps)
    from glycresoft.scoring.elution_time_grouping import GlycopeptideChromatogramProxy
    cases = []
    analysis_name = analysis.name
    start_time, stop_time = apex_time_range
    for i, idgp in enumerate(idgps):
        if i % 50 == 0:
            click.echo("%d/%d Records Processed" % (i, n), err=True)
        if idgp.chromatogram is None:
            continue
        if idgp.ms1_score < 0:
            continue
        if idgp.q_value > fdr_threshold:
            continue
        obj = GlycopeptideChromatogramProxy.from_chromatogram(
            idgp, ms1_score=idgp.ms1_score, ms2_score=idgp.ms2_score,
            q_value=idgp.q_value, analysis_name=analysis_name,
            mass_shifts=idgp.chromatogram.mass_shifts)
        if obj.apex_time < start_time or obj.apex_time > stop_time:
            continue
        cases.append(obj)
    click.echo("Writing %d Records" % (len(cases), ), err=True)
    with fh:
        GlycopeptideChromatogramProxy.to_csv(cases, csv_stream(fh))



import time

from queue import Empty, Full, Queue
from typing import List, Optional, Tuple, Type, Union, Dict, Set, TYPE_CHECKING

from glycresoft.chromatogram_tree.mass_shift import MassShiftBase

from ms_deisotope.data_source import ProcessedScan, ProcessedRandomAccessScanSource
# from ms_deisotope.output import ProcessedMSFileLoader

from glycresoft.structure.scan import ScanStub
from glycresoft.tandem.glycopeptide.scoring.base import GlycopeptideSpectrumMatcherBase

from glycresoft.task import TaskExecutionSequence
from glycresoft.chromatogram_tree import Unmodified

from .search_space import (
    Parser,
    PredictiveGlycopeptideSearch)

from ...workload import WorkloadManager
from ...spectrum_match.solution_set import MultiScoreSpectrumSolutionSet

from ..scoring import LogIntensityScorer
from ..matcher import GlycopeptideMatcher

if TYPE_CHECKING:
    from multiprocessing.managers import SyncManager
    from multiprocessing.synchronize import Event as MPEvent

Full: Exception

class MultiScoreGlycopeptideMatcher(GlycopeptideMatcher):
    solution_set_type = MultiScoreSpectrumSolutionSet


def IsTask(cls):
    return cls


def workload_grouping(
        chunks: List[List[ProcessedScan]],
        max_scans_per_workload: int=500,
        starting_index: int=0) -> Tuple[List[List[ProcessedScan]], int]:
    """
    Gather together precursor mass batches of tandem mass spectra into a unit
    batch, starting from index ``starting_index``.

    Parameters
    ----------
    chunks : List[List[:class:`~.ProcessedScan`]]
        The precursor mass batched spectra
    max_scans_per_workload : int
        The total number of scans to include in a unit batch. This is a soft
        upper bound, a unit batch may have more than this if the next precursor
        mass batch is too large, but no additional batches will be added after
        that.
    starting_index : int
        The offset into ``chunks`` to start collecting batches from

    Returns
    -------
    unit_batch : List[List[:class:`~.ProcessedScan`]]
        The batch of precursor mass batches
    ending_index : int
        The offset into ``chunks`` that the next round should
        start from.
    """
    workload = []
    total_scans_in_workload = 0
    i = starting_index
    n = len(chunks)
    while total_scans_in_workload < max_scans_per_workload and i < n:
        chunk = chunks[i]
        workload.append(chunk)
        total_scans_in_workload += len(chunk)
        i += 1
    return workload, i


class SpectrumBatcher(TaskExecutionSequence):
    """
    Break a big list of scans into precursor mass batched blocks of
    spectrum groups with an approximate maximum size. Feeds raw workloads
    into the pipeline.
    """
    groups: List
    out_queue: Queue
    max_scans_per_workload: int

    def __init__(self, groups, out_queue, max_scans_per_workload=250):
        self.groups = groups
        self.max_scans_per_workload = max_scans_per_workload
        self.out_queue = out_queue
        self.done_event = self._make_event()

    def generate(self):
        groups = self.groups
        max_scans_per_workload = self.max_scans_per_workload
        group_n = len(groups)
        group_i = 0
        while group_i < group_n:
            group_i_prev = group_i
            chunk, group_i = workload_grouping(groups, max_scans_per_workload, group_i)
            yield chunk, group_i_prev, group_n

    def run(self):
        for batch in self.generate():
            if self.error_occurred():
                break
            while not self.error_occurred():
                try:
                    self.out_queue.put(batch, True, 5)
                    break
                except Full:
                    pass
        self.done_event.set()


class BatchMapper(TaskExecutionSequence):
    """
    Wrap scan groups from :class:`SpectrumBatcher` in :class:`StructureMapper` tasks
    and ships them to an appropriate work queue (usually another process).

    .. note::
        The StructureMapper could be applied with or without a database bound to it,
        and for an IPC consumer the database should not be bound, only labeled.
    """

    search_groups: List[List[ProcessedScan]]
    precursor_error_tolerance: float
    mass_shifts: List[MassShiftBase]

    in_queue: Queue
    out_queue: Dict[str, Queue]
    in_done_event: 'MPEvent'
    done_event: 'MPEvent'

    def __init__(self, search_groups, in_queue, out_queue, in_done_event,
                 precursor_error_tolerance=5e-6, mass_shifts=None):
        if mass_shifts is None:
            mass_shifts = [Unmodified]
        if not isinstance(out_queue, dict):
            for label, _group in search_groups:
                out_queue = {label: out_queue}
        self.search_groups = search_groups
        self.precursor_error_tolerance = precursor_error_tolerance
        self.mass_shifts = mass_shifts
        self.in_queue = in_queue
        self.out_queue = out_queue
        self.in_done_event = in_done_event
        self.done_event = self._make_event()

    def out_queue_for_label(self, label):
        return self.out_queue[label]

    def execute_task(self, task: Tuple[List[List[ProcessedScan]], int, int]):
        chunk, group_i_prev, group_n = task
        for label, search_group in self.search_groups:
            task = StructureMapper(
                chunk, group_i_prev, group_n, search_group,
                precursor_error_tolerance=self.precursor_error_tolerance,
                mass_shifts=self.mass_shifts)
            task.label = label
            # Introduces a thread safety issue?
            task.unbind_scans()
            while not self.error_occurred():
                try:
                    self.out_queue_for_label(label).put(task, True, 5)
                    break
                except Full:
                    pass

    def run(self):
        has_work = True
        while has_work and not self.error_occurred():
            try:
                task = self.in_queue.get(True, 5)
                self.execute_task(task)
            except Empty:
                if self.in_done_event.is_set():
                    has_work = False
                    break
        self.done_event.set()


@IsTask
class StructureMapper(TaskExecutionSequence[WorkloadManager]):
    """
    Map spectra against the database using a precursor filtering search strategy,
    generating a task graph of spectrum-structure-mass_shift relationships, a
    :class:`~.WorkloadManager` instance.
    """

    chunk: List[List[ProcessedScan]]
    group_i: int
    group_n: int
    predictive_search: Union[PredictiveGlycopeptideSearch, str]
    precursor_error_tolerance: float
    mass_shifts: List[MassShiftBase]
    seen: Set[str]

    def __init__(self, chunk, group_i, group_n, predictive_search, precursor_error_tolerance=5e-6,
                 mass_shifts=None):
        if mass_shifts is None:
            mass_shifts = [Unmodified]
        self.chunk = chunk
        self.group_i = group_i
        self.group_n = group_n
        self.predictive_search = predictive_search
        self.seen = set()
        self.mass_shifts = mass_shifts
        self.precursor_error_tolerance = precursor_error_tolerance

    def bind_scans(self, source: ProcessedRandomAccessScanSource):
        for group in self.chunk:
            for scan in group:
                scan.bind(source)

    def unbind_scans(self):
        for group in self.chunk:
            for scan in group:
                scan.unbind()

    def get_scan_source(self):
        for group in self.chunk:
            for scan in group:
                return scan.source

    def _log_cache(self):
        if False:
            predictive_search = self.predictive_search
            hits = predictive_search.peptide_glycosylator._cache_hit
            misses = predictive_search.peptide_glycosylator._cache_miss
            total = hits + misses
            if total > 5000:
                self.log("... Cache Performance: %d / %d (%0.2f%%)" % (hits, total, hits / float(total) * 100.0))

    def _prepare_scan(self, scan: Union[ScanStub, ProcessedScan]) -> ProcessedScan:
        try:
            return scan.convert()
        except AttributeError:
            if isinstance(scan, ProcessedScan):
                return scan
            else:
                raise

    def map_structures(self) -> WorkloadManager:
        counter = 0
        workload = WorkloadManager()
        predictive_search = self.predictive_search
        start = time.time()
        total_work = 0
        lo_min = float('inf')
        hi_max = 0
        for i, group in enumerate(self.chunk):
            lo = float('inf')
            hi = 0
            temp = []
            for g in group:
                g = self._prepare_scan(g)
                if g.id in self.seen:
                    raise ValueError("Repeated Scan %r" % g.id)
                self.seen.add(g.id)
                counter += 1
                mass = g.precursor_information.neutral_mass
                temp.append(g)
                if lo > mass:
                    lo = mass
                if hi < mass:
                    hi = mass
            group = temp
            solutions = predictive_search.handle_scan_group(
                group, mass_shifts=self.mass_shifts, precursor_error_tolerance=self.precursor_error_tolerance)
            total_work += solutions.total_work_required()
            if i % 500 == 0 and i != 0:
                self.log('... Mapped Group %d (%0.2f%%) %0.3f-%0.3f with %d Items (%d Total)' % (
                    i + self.group_i, i * 100.0 / len(self.chunk), lo, hi,
                    solutions.total_work_required(), total_work))
            lo_min = min(lo_min, lo)
            hi_max = max(hi_max, hi)
            workload.update(solutions)
        end = time.time()
        if counter:
            self.debug("... Mapping Completed %0.3f-%0.3f (%0.2f sec.)" %
                       (lo_min, hi_max, end - start))
        self._log_cache()
        predictive_search.reset()
        return workload

    def add_decoy_glycans(self, workload: WorkloadManager) -> WorkloadManager:
        items = list(workload.hit_map.items())
        for hit_id, record in items:
            record = record.to_decoy_glycan()
            for scan in workload.hit_to_scan_map[hit_id]:
                hit_type = workload.scan_hit_type_map[scan.id, hit_id]
                workload.add_scan_hit(scan, record, hit_type)
        return workload

    def run(self) -> WorkloadManager:
        workload = self.map_structures()
        workload.pack()
        self.add_decoy_glycans(workload)
        self.predictive_search.construct_peptide_groups(workload)
        return workload


class MapperExecutor(TaskExecutionSequence):
    """This task executor consumes batches of precursor mass-grouped spectra,
    and produces batches of glycopeptides matched to spectra.

    Its task type is :class:`StructureMapper`

    """

    scan_loader: ProcessedRandomAccessScanSource
    predictive_searchers: Dict[str, PredictiveGlycopeptideSearch]

    in_queue: Queue
    out_queue: Queue
    in_done_event: "MPEvent"
    done_event: "MPEvent"

    def __init__(self, predictive_searchers, scan_loader, in_queue, out_queue, in_done_event):
        self.in_queue = in_queue
        self.out_queue = out_queue
        self.in_done_event = in_done_event
        self.done_event = self._make_event()
        self.scan_loader = scan_loader
        self.predictive_searchers = predictive_searchers

    def execute_task(self, mapper_task: StructureMapper) -> 'SpectrumMatcher':
        self.scan_loader.reset()
        # In case this came from a labeled batch mapper and not
        # attached to an actual dynamic glycopeptide generator
        batch_label = None
        label = mapper_task.predictive_search
        if isinstance(label, str):
            batch_label = label
            mapper_task.predictive_search = self.predictive_searchers[label]

        mapper_task.bind_scans(self.scan_loader)
        workload = mapper_task()
        self.scan_loader.reset()
        matcher_task = SpectrumMatcher(
            workload,
            mapper_task.group_i,
            mapper_task.group_n,
            batch_label=batch_label)
        return matcher_task

    def run(self):
        has_work = True
        strikes = 0
        while has_work and not self.error_occurred():
            try:
                mapper_task = self.in_queue.get(True, 5)
                matcher_task = self.execute_task(mapper_task)
                while not self.error_occurred():
                    try:
                        self.out_queue.put(matcher_task, True, 5)
                        break
                    except Full:
                        pass
                # Detach the scans from the scan source again.
                mapper_task.unbind_scans()

            except Empty:
                strikes += 1
                if strikes % 50 == 0 and strikes:
                    self.log("... %d iterations without new batches on %s, done event: %s" % (
                        strikes, self, self.in_done_event.is_set()))
                if self.in_done_event.is_set():
                    has_work = False
                    break
        self.done_event.set()


class SemaphoreBoundMapperExecutor(MapperExecutor):
    def __init__(self, semaphore, predictive_searchers, scan_loader, in_queue, out_queue,
                 in_done_event, tracking_directory=None):
        super(SemaphoreBoundMapperExecutor, self).__init__(
            predictive_searchers, scan_loader,
            in_queue, out_queue, in_done_event)
        self.semaphore = semaphore

    def execute_task(self, mapper_task):
        with self.semaphore:
            result = super(SemaphoreBoundMapperExecutor, self).execute_task(mapper_task)
        return result


@IsTask
class SpectrumMatcher(TaskExecutionSequence[List[MultiScoreSpectrumSolutionSet]]):
    """Actually execute the spectrum matching specified in a :class:`~.WorkloadManager`.

    .. note::
        This task may spin up additional processes if :attr:`n_processes` is greater than
        1, but it must be ~4 or better usually to have an appreciable speedup compared to
        a executing the matching in serial. IPC communication is expensive, no matter what.
    """

    workload: WorkloadManager
    group_i: int
    group_n: int
    scorer_type: Type[GlycopeptideSpectrumMatcherBase]
    ipc_manager: 'SyncManager'
    batch_label: Optional[str]

    n_processes: int
    mass_shifts: List[MassShiftBase]
    evaluation_kwargs: Dict
    cache_seeds: Optional[Dict]

    def __init__(self, workload, group_i, group_n, scorer_type=None,
                 ipc_manager=None, n_processes=6, mass_shifts=None,
                 evaluation_kwargs=None, cache_seeds=None,
                 batch_label: Optional[str]=None,
                 **kwargs):
        if scorer_type is None:
            scorer_type = LogIntensityScorer
        if evaluation_kwargs is None:
            evaluation_kwargs = {}
        self.workload = workload
        self.group_i = group_i
        self.group_n = group_n
        self.batch_label = batch_label

        self.mass_shifts = mass_shifts
        self.scorer_type = scorer_type
        self.evaluation_kwargs = evaluation_kwargs
        self.evaluation_kwargs.update(kwargs)

        self.ipc_manager = ipc_manager
        self.n_processes = n_processes
        self.cache_seeds = cache_seeds

    def score_spectra(self) -> List[MultiScoreSpectrumSolutionSet]:
        matcher = MultiScoreGlycopeptideMatcher(
            [], self.scorer_type, None, Parser,
            ipc_manager=self.ipc_manager,
            n_processes=self.n_processes,
            mass_shifts=self.mass_shifts,
            cache_seeds=self.cache_seeds)

        target_solutions = []
        lo, hi = self.workload.mass_range()
        batches = list(self.workload.batches(matcher.batch_size))
        if self.workload.total_size:
            label = ''
            if self.batch_label:
                label = self.batch_label.title() + ' '
            self.log(
                f"... {label}{max((self.group_i - 1), 0) * 100.0 / self.group_n:0.2f}% "
                f"({lo:0.2f}-{hi:0.2f}) {self.workload}")


        n_batches = len(batches)
        running_total_work = 0
        total_work = self.workload.total_work_required()
        self.workload.clear()
        for i, batch in enumerate(batches):
            if batch.batch_size == 0:
                batch.clear()
                continue
            if n_batches > 1:
                self.log("... Batch %d (%d/%d) %0.2f%%" % (
                    i + 1, running_total_work + batch.batch_size, total_work,
                    ((running_total_work + batch.batch_size) * 100.) / float(total_work + 1)))
            running_total_work += batch.batch_size
            target_scan_solution_map = matcher.evaluate_hit_groups(
                batch, **self.evaluation_kwargs)

            temp = matcher.collect_scan_solutions(
                target_scan_solution_map, batch.scan_map)
            temp = [case for case in temp if len(case) > 0]
            for case in temp:
                case.simplify()
                # Don't run the select top filters for consistency. They seemed to
                # influence reproducibility.
            target_solutions.extend(temp)
            batch.clear()

        if batches:
            label = ''
            if self.batch_label:
                label = self.batch_label.title() + ' '
            self.log(f"... Finished {label}{max(self.group_i - 1, 0) * 100.0 / self.group_n:0.2f}%"
                     f" ({lo:0.2f}-{hi:0.2f})")
        return target_solutions

    def run(self):
        solution_sets = self.score_spectra()
        return solution_sets


class MatcherExecutor(TaskExecutionSequence):
    """This task executor consumes mappings from glycopeptide to scan and runs spectrum
    matching, scoring each glycopeptide against their matched spectra. It produces  scored
    spectrum matches.

    This type complements :class:`MapperExecutor`

    Its task type is :class:`SpectrumMatcher`
    """

    in_queue: Queue
    out_queue: Queue
    in_done_event: 'MPEvent'
    done_event: 'MPEvent'
    ipc_manager: 'SyncManager'

    scorer_type: Type[GlycopeptideSpectrumMatcherBase]

    n_processes: int
    mass_shifts: List[MassShiftBase]
    evaluation_kwargs: Dict
    cache_seeds: Optional[Dict]

    def __init__(self, in_queue, out_queue, in_done_event, scorer_type=None, ipc_manager=None,
                 n_processes=6, mass_shifts=None, evaluation_kwargs=None, cache_seeds=None,
                 **kwargs):
        if scorer_type is None:
            scorer_type = LogIntensityScorer
        if evaluation_kwargs is None:
            evaluation_kwargs = {}

        self.in_queue = in_queue
        self.out_queue = out_queue
        self.in_done_event = in_done_event
        self.done_event = self._make_event()

        self.mass_shifts = mass_shifts
        self.scorer_type = scorer_type
        self.evaluation_kwargs = evaluation_kwargs
        self.evaluation_kwargs.update(kwargs)

        self.n_processes = n_processes
        self.ipc_manager = ipc_manager
        self.cache_seeds = cache_seeds

    def configure_task(self, matcher_task: SpectrumMatcher):
        matcher_task.ipc_manager = self.ipc_manager
        matcher_task.n_processes = self.n_processes
        matcher_task.scorer_type = self.scorer_type
        matcher_task.evaluation_kwargs = self.evaluation_kwargs
        matcher_task.mass_shifts = self.mass_shifts
        matcher_task.mass_shift_map = {m.name: m for m in self.mass_shifts}
        return matcher_task

    def execute_task(self, matcher_task):
        matcher_task = self.configure_task(matcher_task)
        solutions = matcher_task()
        return solutions

    def run(self):
        has_work = True
        while has_work and not self.error_occurred():
            try:
                matcher_task: SpectrumMatcher = self.in_queue.get(True, 3)
                solutions = self.execute_task(matcher_task)
                while not self.error_occurred():
                    try:
                        self.out_queue.put(solutions, True, 5)
                        break
                    except Full:
                        pass
            except Empty:
                if self.in_done_event.is_set():
                    has_work = False
                    break
        self.done_event.set()


class SemaphoreBoundMatcherExecutor(MatcherExecutor):
    def __init__(self, semaphore, in_queue, out_queue, in_done_event, scorer_type=None,
                 ipc_manager=None, n_processes=6, mass_shifts=None, evaluation_kwargs=None,
                 cache_seeds=None, **kwargs):
        super(SemaphoreBoundMatcherExecutor, self).__init__(
            in_queue, out_queue, in_done_event, scorer_type, ipc_manager,
            n_processes, mass_shifts, evaluation_kwargs, cache_seeds=cache_seeds, **kwargs)
        self.semaphore = semaphore

    def execute_task(self, matcher_task):
        with self.semaphore:
            result = super(SemaphoreBoundMatcherExecutor, self).execute_task(matcher_task)
        return result



import sys
import multiprocessing
import click
import textwrap

from glycresoft.cli.base import cli, HiddenOption

from glycresoft.cli.validators import (
    glycan_source_validators,
    validate_modifications,
    validate_glycan_source,
    validate_glycopeptide_hypothesis_name,
    validate_glycan_hypothesis_name,
    get_by_name_or_id,
    validate_reduction,
    validate_derivatization,
    validate_mzid_proteins,
    GlycanHypothesisCopier,
    DatabaseConnectionParam,
    SubstituentParamType)

from glycresoft.cli.utils import ctxstream

from glycresoft.serialize import (
    DatabaseBoundOperation,
    GlycanHypothesis,
    Analysis,
    AnalysisTypeEnum)

from glycresoft.database.builder.glycopeptide.naive_glycopeptide import (
    MultipleProcessFastaGlycopeptideHypothesisSerializer,
    ReversingMultipleProcessFastaGlycopeptideHypothesisSerializer,
    NonSavingMultipleProcessFastaGlycopeptideHypothesisSerializer)

from glycresoft.database.builder.glycopeptide.informed_glycopeptide import (
    MultipleProcessMzIdentMLGlycopeptideHypothesisSerializer,
    ReversingMultipleProcessMzIdentMLGlycopeptideHypothesisSerializer)

from glycresoft.database.builder.glycan import (
    TextFileGlycanHypothesisSerializer,
    CombinatorialGlycanHypothesisSerializer,
    GlycanCompositionHypothesisMerger,
    NGlycanGlyspaceHypothesisSerializer,
    OGlycanGlyspaceHypothesisSerializer,
    TaxonomyFilter,
    GlycanAnalysisHypothesisSerializer,
    GlycopeptideAnalysisGlycanCompositionExtractionHypothesisSerializer,
    SynthesisGlycanHypothesisSerializer)

from glycresoft.database.prebuilt import hypothesis_register as prebuilt_hypothesis_register

from glycresoft.database.disk_backed_database import GlycanCompositionDiskBackedStructureDatabase
from glycresoft.database.composition_network import (
    CompositionGraph, GraphWriter, GraphReader,
    CompositionRangeRule,
    CompositionRatioRule,
    CompositionRuleClassifier,
    CompositionExpressionRule,
    make_n_glycan_neighborhoods,
    make_adjacency_neighborhoods,
    make_mammalian_n_glycan_neighborhoods)

from glycresoft.database.builder.glycan.synthesis import synthesis_register

from glycopeptidepy.enzyme import enzyme_rules
from glycopeptidepy.utils.collectiontools import decoratordict
from glycopeptidepy.structure.modification import RestrictedModificationTable


@cli.group('build-hypothesis', short_help='Build search spaces for glycans and glycopeptides')
def build_hypothesis():
    pass


command_group = build_hypothesis


def database_connection(fn):
    arg = click.argument("database-connection", doc_help=(
        "A connection URI for a database, or a path on the file system"))
    return arg(fn)


_glycan_hypothesis_builders = decoratordict()


@_glycan_hypothesis_builders('text')
def _build_glycan_composition_hypothesis_from_text(database_connection, text_file, hypothesis_name,
                                                   identifier=None):
    if hypothesis_name is not None:
        hypothesis_name = hypothesis_name + '-Glycans'
    builder = TextFileGlycanHypothesisSerializer(
        text_file, database_connection, hypothesis_name=hypothesis_name)
    builder.run()
    return builder.hypothesis_id


@_glycan_hypothesis_builders("combinatorial")
def _build_glycan_composition_hypothesis_from_combinatorial(database_connection, text_file, hypothesis_name,
                                                            identifier=None):
    if hypothesis_name is not None:
        hypothesis_name = hypothesis_name + '-Glycans'
    builder = CombinatorialGlycanHypothesisSerializer(
        text_file, database_connection, hypothesis_name=hypothesis_name)
    builder.run()
    return builder.hypothesis_id


@_glycan_hypothesis_builders("hypothesis")
def _copy_hypothesis_across_file_boundaries(database_connection, source, hypothesis_name,
                                            identifier=None):
    source_handle = DatabaseBoundOperation(source)
    source_hypothesis_id = None
    source_hypothesis_name = None

    try:
        hypothesis_id = int(identifier)
        inst = source_handle.query(GlycanHypothesis).get(hypothesis_id)
        if inst is not None:
            source_hypothesis_id = hypothesis_id
            source_hypothesis_name = inst.name

    except TypeError:
        hypothesis_name = identifier
        inst = source_handle.query(GlycanHypothesis).filter(
            GlycanHypothesis.name == hypothesis_name).first()
        if inst is not None:
            source_hypothesis_id = inst.id
            source_hypothesis_name = inst.name

    if source == database_connection:
        return source_hypothesis_id

    mover = GlycanHypothesisCopier(
        database_connection, [(source, source_hypothesis_id)],
        hypothesis_name=source_hypothesis_name)
    mover.run()
    return mover.hypothesis_id


@_glycan_hypothesis_builders("analysis")
def _copy_analysis_across_file_boundaries(database_connection, source, hypothesis_name,
                                          identifier=None):
    source_handle = DatabaseBoundOperation(source)
    source_analysis_id = None
    source_analysis_name = None
    try:
        hypothesis_id = int(identifier)
        inst = source_handle.query(Analysis).get(hypothesis_id)
        if inst is not None:
            source_analysis_id = hypothesis_id
            source_analysis_name = inst.name

    except TypeError:
        hypothesis_name = identifier
        inst = source_handle.query(Analysis).filter(
            Analysis.name == hypothesis_name).first()
        if inst is not None:
            source_analysis_id = inst.id
            source_analysis_name = inst.name
    if hypothesis_name is None:
        hypothesis_name = source_analysis_name
    mover = GlycanAnalysisHypothesisSerializer(
        source, source_analysis_id, hypothesis_name,
        database_connection)
    mover.run()
    return mover.hypothesis_id


def _validate_glycan_source_identifier(ctx, param, value):
    try:
        if ctx.params['glycan_source_type'] not in ("hypothesis", "analysis") and value:
            width = click.get_terminal_size()[0]
            click.secho('\n'.join(
                textwrap.wrap(
                    "Warning: --glycan-source-identifier specified when "
                    "--glycan-source is neither \"hypothesis\" nor \"analysis\"."
                    " Its value will be ignored.", width=int(width * 0.6))), fg='yellow')
            return None
        else:
            return value
    except KeyError:
        click.secho("Specify --glycan-source before --glycan-source-identifier.", fg='yellow')


def glycopeptide_hypothesis_common_options(cmd):
    options = [
        click.option("-u", "--occupied-glycosites", type=int, default=1,
                     help=("The number of occupied glycosylation sites permitted.")),
        click.option("-n", "--name", default=None, help="The name for the hypothesis to be created"),
        click.option("-p", "--processes", 'processes', type=click.IntRange(1, multiprocessing.cpu_count()),
                     default=min(multiprocessing.cpu_count(), 4),
                     help=('Number of worker processes to use. Defaults to 4 '
                           'or the number of CPUs, whichever is lower')),
        click.option("-G", "--glycan-source-identifier", required=False, default=None,
                     help=("The name or id number of the hypothesis or analysis to"
                           " be used when using those glycan source types."),
                     callback=_validate_glycan_source_identifier),
        click.option("-s", "--glycan-source-type", default='text', type=click.Choice(
                     list(glycan_source_validators.keys())),
                     help="The type of glycan information source to use"),
        click.option("-g", "--glycan-source", required=True,
                     help="The path, identity, or other specifier for the glycan source"),
        click.option("-z", "--peptide-length-range", type=(int, int), default=(5, 60),
                     help="The minimum and maximum peptide length to consider")
    ]
    for opt in options:
        cmd = opt(cmd)
    return cmd


@build_hypothesis.command("glycopeptide-fa",
                          short_help="Build glycopeptide search spaces with a FASTA file of proteins")
@click.pass_context
@glycopeptide_hypothesis_common_options
@click.argument("fasta-file", type=click.Path(exists=True), doc_help=(
    "A file containing protein sequences in FASTA format"))
@database_connection
@click.option("-e", "--enzyme", default=['trypsin'], multiple=True,
              help=(
                  'The proteolytic enzyme to use during digestion. May be specified multiple'
                  ' times, generating a co-digestion. May specify an enzyme name or a regular'
                  ' expression describing the cleavage pattern. Recognized enzyme names are: ' +
                  ', '.join(sorted(enzyme_rules))))
@click.option("-m", "--missed-cleavages", type=int, default=1,
              help="The number of missed proteolytic cleavage sites permitted")
@click.option("-c", "--constant-modification", multiple=True,
              help='Peptide modification rule which will be applied constantly')
@click.option("-v", "--variable-modification", multiple=True,
              help='Peptide modification rule which will be applied variablely')
@click.option("-V", "--max-variable-modifications", type=int, default=4, required=False,
              help=("The maximum number of variable modifications that can be applied to a single peptide"))
@click.option("-y", "--semispecific-digest", is_flag=True, help=(
    "Apply a semispecific enzyme digest permitting one peptide terminal to be non-specific"))
@click.option("-R", "--reverse", default=False, is_flag=True, help='Reverse protein sequences')
@click.option("--dry-run", default=False, is_flag=True, help="Do not save glycopeptides", cls=HiddenOption)
@click.option("-F", "--not-full-crossproduct", is_flag=True, help=(
    "Do not produce full crossproduct. For when the search space is too large to enumerate, store, and load."))
@click.option("--retain-all-peptides", is_flag=True, default=False,
              help=("Do not require a glycosylation site when saving base peptides"))
@click.option("-C", "--include-n-x-c-glycosylation", "include_cysteine_n_glycosylation", default=False, is_flag=True,
              help="Whether to support N-x-C in the N-glycosylation sequon")
@click.option("-a", "--annotation-path", type=click.Path(exists=True, allow_dash=True), required=False,
              default=None, help="The path to an annotation XML file from UniProt, or '-' to suppress annotations.")
def glycopeptide_fa(context, fasta_file, database_connection, enzyme, missed_cleavages, occupied_glycosites, name,
                    constant_modification, variable_modification, processes, glycan_source, glycan_source_type,
                    glycan_source_identifier=None, semispecific_digest=False, reverse=False, dry_run=False,
                    peptide_length_range=(5, 60), not_full_crossproduct=False, max_variable_modifications=4,
                    retain_all_peptides=False, include_cysteine_n_glycosylation: bool=False, annotation_path=None):
    '''Constructs a glycopeptide hypothesis from a FASTA file of proteins and a
    collection of glycans.
    '''
    if reverse:
        task_type = ReversingMultipleProcessFastaGlycopeptideHypothesisSerializer
        click.secho("Using ReversingMultipleProcessFastaGlycopeptideHypothesisSerializer", fg='yellow')
    elif dry_run:
        task_type = NonSavingMultipleProcessFastaGlycopeptideHypothesisSerializer
        click.secho("Using NonSavingMultipleProcessFastaGlycopeptideHypothesisSerializer", fg='yellow')
    else:
        task_type = MultipleProcessFastaGlycopeptideHypothesisSerializer

    validate_modifications(
        context, constant_modification + variable_modification)
    validate_glycan_source(context, database_connection,
                           glycan_source, glycan_source_type,
                           glycan_source_identifier)

    processes = min(multiprocessing.cpu_count(), processes)

    if name is not None:
        name = validate_glycopeptide_hypothesis_name(
            context, database_connection, name)
        click.secho("Building Glycopeptide Hypothesis %s" % name, fg='cyan')
    mt = RestrictedModificationTable(
        None, constant_modification, variable_modification)
    constant_modification = [mt[c] for c in constant_modification]
    variable_modification = [mt[c] for c in variable_modification]

    glycan_hypothesis_id = _glycan_hypothesis_builders[
        glycan_source_type](database_connection, glycan_source, name, glycan_source_identifier)

    builder = task_type(
        fasta_file, database_connection,
        glycan_hypothesis_id=glycan_hypothesis_id,
        protease=enzyme,
        constant_modifications=constant_modification,
        variable_modifications=variable_modification,
        max_missed_cleavages=missed_cleavages,
        max_glycosylation_events=occupied_glycosites,
        hypothesis_name=name,
        semispecific=semispecific_digest,
        n_processes=processes,
        full_cross_product=not not_full_crossproduct,
        max_variable_modifications=max_variable_modifications,
        peptide_length_range=peptide_length_range,
        require_glycosylation_sites=not retain_all_peptides,
        include_cysteine_n_glycosylation=include_cysteine_n_glycosylation,
        uniprot_source_file=annotation_path)
    builder.display_header()
    builder.start()
    return builder.hypothesis_id


@build_hypothesis.command("glycopeptide-mzid",
                          short_help="Build a glycopeptide search space with an mzIdentML file")
@click.pass_context
@click.argument("mzid-file", type=click.Path(exists=True))
@database_connection
@glycopeptide_hypothesis_common_options
@click.option("-t", "--target-protein", multiple=True,
              help='Specifies the name of a protein to include in the hypothesis, or a file containing such'
                   ', one per line. May be used many times.')
@click.option('-r', '--target-protein-re', multiple=True,
              help=('Specifies a regular expression to select proteins'
                    ' to be included by name. May be used many times.'))
@click.option("--reference-fasta", default=None, required=False,
              help=("When the full sequence for each protein is not embedded in the mzIdentML file and "
                    "the FASTA file used is not local."))
@click.option("-a", "--annotation-path", type=click.Path(exists=True, allow_dash=True), required=False,
              default=None, help="The path to an annotation XML file from UniProt or '-' to suppress loading annotations")
@click.option("-R", "--reverse", default=False, is_flag=True, help='Reverse protein sequences')
@click.option("-F", "--not-full-crossproduct", is_flag=True, help=(
    "Do not produce full crossproduct. For when the search space is too large to enumerate, store, and load."))
def glycopeptide_mzid(context, mzid_file, database_connection, name, occupied_glycosites, target_protein,
                      target_protein_re, processes, glycan_source, glycan_source_type, glycan_source_identifier,
                      reference_fasta, peptide_length_range=(5, 60), annotation_path=None, reverse=False,
                      not_full_crossproduct=False):
    """
    Constructs a glycopeptide hypothesis from a MzIdentML file of proteins and a
    collection of glycans.
    """
    proteins = validate_mzid_proteins(
        context, mzid_file, target_protein, target_protein_re)
    validate_glycan_source(context, database_connection,
                           glycan_source, glycan_source_type,
                           glycan_source_identifier)

    processes = min(multiprocessing.cpu_count(), processes)

    if name is not None:
        name = validate_glycopeptide_hypothesis_name(
            context, database_connection, name)
        click.secho("Building Glycopeptide Hypothesis %s" % name, fg='cyan')

    glycan_hypothesis_id = _glycan_hypothesis_builders[
        glycan_source_type](database_connection, glycan_source, name, glycan_source_identifier)

    job_cls = MultipleProcessMzIdentMLGlycopeptideHypothesisSerializer
    if reverse:
        job_cls = ReversingMultipleProcessMzIdentMLGlycopeptideHypothesisSerializer

    builder = job_cls(
        mzid_file, database_connection,
        glycan_hypothesis_id=glycan_hypothesis_id,
        hypothesis_name=name,
        target_proteins=proteins,
        max_glycosylation_events=occupied_glycosites,
        reference_fasta=reference_fasta,
        n_processes=processes,
        peptide_length_range=peptide_length_range,
        uniprot_source_file=annotation_path,
        full_cross_product=not not_full_crossproduct
    )
    builder.display_header()
    builder.start()
    return builder.hypothesis_id


@build_hypothesis.command("glycan-text",
                          short_help='Build a glycan search space with a text file of glycan compositions')
@click.pass_context
@click.argument("text-file", type=click.Path(exists=True))
@database_connection
@click.option("-r", "--reduction", default=None, help='Reducing end modification')
@click.option("-d", "--derivatization", default=None, help='Chemical derivatization to apply')
@click.option("-n", "--name", default=None, help="The name for the hypothesis to be created")
def glycan_text(context, text_file, database_connection, reduction, derivatization, name):
    if name is not None:
        name = validate_glycan_hypothesis_name(context, database_connection, name)
        click.secho("Building Glycan Hypothesis %s" % name, fg='cyan')
    reduction = validate_reduction(context, reduction)
    derivatization = validate_derivatization(context, derivatization)
    builder = TextFileGlycanHypothesisSerializer(
        text_file, database_connection, reduction=reduction, derivatization=derivatization,
        hypothesis_name=name)
    builder.display_header()
    builder.start()


@build_hypothesis.command("glycan-combinatorial", short_help=('Build a glycan search space with a text file'
                                                              ' containing algebraic combination rules'))
@click.pass_context
@click.argument("rule-file", type=click.Path(exists=True))
@database_connection
@click.option("-r", "--reduction", default=None, help='Reducing end modification')
@click.option("-d", "--derivatization", default=None, help='Chemical derivatization to apply')
@click.option("-n", "--name", default=None, help="The name for the hypothesis to be created")
@click.option("-c", "--glycan-type", default=None, required=False,
              type=click.Choice(['n-glycan', 'o-glycan', 'gag-linker'], case_sensitive=False),
              help="The glycan classification to apply to all generated glycan compositions."
                   " If not set, a heuristic will be used to guess if a composition in an N-glycan."
                   " This is only relevant when building a glycopeptide hypothesis.")
def glycan_combinatorial(context, rule_file, database_connection, reduction, derivatization, name,
                         glycan_type=None):
    if name is not None:
        name = validate_glycan_hypothesis_name(context, database_connection, name)
        click.secho("Building Glycan Hypothesis %s" % name, fg='cyan')
    reduction = validate_reduction(context, reduction)
    derivatization = validate_derivatization(context, derivatization)
    builder = CombinatorialGlycanHypothesisSerializer(
        rule_file, database_connection, reduction=reduction, derivatization=derivatization,
        hypothesis_name=name, glycan_type=glycan_type)
    builder.display_header()
    builder.start()


@build_hypothesis.command("merge-glycan", short_help=("Combine two or more glycan search spaces to create a "
                                                      "new one containing unique entries from all constituents"))
@click.pass_context
@database_connection
@click.option("-n", "--name", default=None, help="The name for the hypothesis to be created")
@click.option(
    "-i", "--hypothesis-specification", multiple=True,
    nargs=2, help=("The location and identity for the hypothesis to"
                   " include. May be specified many times"))
def merge_glycan_hypotheses(context, database_connection, hypothesis_specification, name):
    database_connection = DatabaseBoundOperation(database_connection)
    hypothesis_ids = []
    for connection, ident in hypothesis_specification:
        hypothesis = get_by_name_or_id(DatabaseBoundOperation(connection), GlycanHypothesis, ident)
        hypothesis_ids.append((connection, hypothesis.id))

    if name is not None:
        name = validate_glycan_hypothesis_name(context, database_connection._original_connection, name)
        click.secho("Building Glycan Hypothesis %s" % name, fg='cyan')

    task = GlycanCompositionHypothesisMerger(
        database_connection._original_connection, hypothesis_ids, name)
    task.display_header()
    task.start()


@build_hypothesis.command("glycan-glyspace", short_help=("Construct a glycan hypothesis from GlySpace"))
@click.pass_context
@database_connection
@click.option("-r", "--reduction", default=None, help='Reducing end modification')
@click.option("-d", "--derivatization", default=None, help='Chemical derivatization to apply')
@click.option("-n", "--name", default=None, help="The name for the hypothesis to be created")
@click.option("-m", "--motif-class", type=click.Choice(["n-linked", "o-linked"]), default=None,
              help="Specify a glycan structure family to search for")
@click.option("-t", "--target-taxon", default=None, help="Only select structures annotated with this taxonomy")
@click.option("-i", "--include-children", default=False, is_flag=True,
              help="Include child taxa of --target-taxon. No effect otherwise.")
@click.option("-s", "--detatch-substituent", multiple=True, type=SubstituentParamType(),
              help='Substituent type to detatch from all monosaccharides')
def glyspace_glycan_hypothesis(context, database_connection, motif_class, reduction, derivatization, name,
                               target_taxon=None, include_children=False, detatch_substituent=None):
    database_connection = DatabaseBoundOperation(database_connection)
    if name is not None:
        name = validate_glycan_hypothesis_name(context, database_connection._original_connection, name)
        click.secho("Building Glycan Hypothesis %s" % name, fg='cyan')
    filter_funcs = []

    if target_taxon is not None:
        filter_funcs.append(TaxonomyFilter(target_taxon, include_children))

    serializer_type = None
    if motif_class == "n-linked":
        serializer_type = NGlycanGlyspaceHypothesisSerializer
    elif motif_class == "o-linked":
        serializer_type = OGlycanGlyspaceHypothesisSerializer

    reduction = validate_reduction(context, reduction)
    derivatization = validate_derivatization(context, derivatization)
    job = serializer_type(
        database_connection._original_connection, name, reduction, derivatization, filter_funcs,
        simplify=True, substituents_to_detatch=detatch_substituent)
    job.display_header()
    job.start()


@build_hypothesis.command("glycan-from-analysis", short_help=("Construct a glycan hypothesis from a matched analysis"))
@click.pass_context
@database_connection
@click.argument("analysis-connection", type=DatabaseConnectionParam(exists=True))
@click.argument("analysis-identifier")
@click.option("-r", "--reduction", default=None, help='Reducing end modification')
@click.option("-d", "--derivatization", default=None, help='Chemical derivatization to apply')
@click.option("-n", "--name", default=None, help="The name for the hypothesis to be created")
def from_analysis(context, database_connection, analysis_connection, analysis_identifier,
                  reduction, derivatization, name):
    database_connection = DatabaseBoundOperation(database_connection)
    if name is not None:
        name = validate_glycan_hypothesis_name(context, database_connection._original_connection, name)
        click.secho("Building Glycan Hypothesis %s" % name, fg='cyan')
    reduction = validate_reduction(context, reduction)
    derivatization = validate_derivatization(context, derivatization)

    analysis_connection = DatabaseBoundOperation(analysis_connection)
    analysis = get_by_name_or_id(analysis_connection.session, Analysis, analysis_identifier)
    if analysis.analysis_type == AnalysisTypeEnum.glycan_lc_ms:
        job = GlycanAnalysisHypothesisSerializer(
            analysis_connection._original_connection, analysis.id, name,
            output_connection=database_connection._original_connection)
        job.display_header()
        job.start()
    elif analysis.analysis_type == AnalysisTypeEnum.glycopeptide_lc_msms:
        job = GlycopeptideAnalysisGlycanCompositionExtractionHypothesisSerializer(
            analysis_connection._original_connection, analysis.id, name,
            output_connection=database_connection._original_connection)
        job.display_header()
        job.start()
    else:
        click.secho("Analysis Type %r could not be converted" % (
            analysis.analysis_type.name,), fg='red')


@build_hypothesis.command("prebuilt-glycan", short_help=(
    'Construct a glycan hypothesis from a list of pre-made recipes'))
@click.pass_context
@database_connection
@click.option("-p", "--recipe-name", type=click.Choice(prebuilt_hypothesis_register.keys()), required=True)
@click.option("-n", "--name", default=None, help="The name for the hypothesis to be created")
@click.option("-r", "--reduction", default=None, help='Reducing end modification')
@click.option("-d", "--derivatization", default=None, help='Chemical derivatization to apply')
def prebuilt_glycan(context, database_connection, recipe_name, name, reduction, derivatization):
    database_connection = DatabaseBoundOperation(database_connection)
    reduction = validate_reduction(context, reduction)
    derivatization = validate_derivatization(context, derivatization)
    if name is not None:
        name = validate_glycan_hypothesis_name(
            context, database_connection._original_connection, name)
    recipe = prebuilt_hypothesis_register[recipe_name]()
    recipe(database_connection._original_connection,
           hypothesis_name=name, reduction=reduction,
           derivatization=derivatization)


@build_hypothesis.command(
    "glycan-synthesis",
    short_help='Construct a glycan hypothesis from a set of biosynthetic pathways')
@click.pass_context
@database_connection
@click.option("-n", "--name", default=None, help="The name for the hypothesis to be created")
@click.option("-r", "--reduction", default=None, help='Reducing end modification')
@click.option("-d", "--derivatization", default=None, help='Chemical derivatization to apply')
@click.option("-w", "--pathway-type", default='human-n-glycan', type=click.Choice([
    'mammalian-n-glycan', 'human-n-glycan',
    'human-o-glycan', 'mammalian-o-glycan']), help=(
    'The default pathway template to use, and the glycan class annotation type.'))
@click.option('-m', '--maximum-size', default=26, type=int,
              help='The maximum number of residues to allow in a single glycan')
@click.option('-e', "--remove-enzyme", "removed_enzymes", multiple=True)
@click.option('-a', "--add-enzyme", "added_enzymes", multiple=True)
@click.option("-p", "--processes", 'processes', type=click.IntRange(1, multiprocessing.cpu_count()),
              default=min(multiprocessing.cpu_count(), 4),
              help=('Number of worker processes to use. Defaults to 4 '
                    'or the number of CPUs, whichever is lower'))
def glycan_synthesis(context, database_connection, name, pathway_type, removed_enzymes, added_enzymes,
                     maximum_size, reduction, derivatization, processes):
    pathway = synthesis_register[pathway_type](n_processes=processes)
    pathway.add_limit(pathway.size_limiter_type(maximum_size))
    for enzyme_to_remove in removed_enzymes:
        pathway.remove_enzyme(enzyme_to_remove)
    for enzyme_to_add in added_enzymes:
        click.echo("Adding enzymes is unsupported at this time")
    task = SynthesisGlycanHypothesisSerializer(
        pathway, database_connection,
        hypothesis_name=name, reduction=reduction, derivatization=derivatization)
    task.start()


@build_hypothesis.group("glycan-network", short_help='Build and manipulate glycan network definitions')
def glycan_network_tools():
    pass


@glycan_network_tools.command("build-network", short_help=(
    "Build a glycan network for an existing glycan hypothesis"))
@click.pass_context
@database_connection
@click.argument("hypothesis-identifier", doc_help=(
                "The ID number or name of the glycan hypothesis to use"))
@click.option("-o", "--output-path", type=click.Path(
    file_okay=True, dir_okay=False, writable=True), default=None,
    help='Path to write to instead of stdout')
@click.option("-e", "--edge-strategy", type=click.Choice(["manhattan", ]), default='manhattan',
              help="Strategy to use to decide when two nodes are connected by an edge")
def glycan_network(context, database_connection, hypothesis_identifier, edge_strategy, output_path):
    conn = DatabaseBoundOperation(database_connection)
    hypothesis = get_by_name_or_id(conn, GlycanHypothesis, hypothesis_identifier)
    if output_path is None:
        output_stream = ctxstream(sys.stdout)
    else:
        output_stream = open(output_path, 'w')
    with output_stream:
        db = GlycanCompositionDiskBackedStructureDatabase(
            database_connection, hypothesis.id)
        glycans = list(db)
        graph = CompositionGraph(glycans)
        if edge_strategy == 'manhattan':
            graph.create_edges(1)
        else:
            raise click.ClickException(
                "Could not find edge strategy %r" % (edge_strategy,))
        GraphWriter(graph, output_stream)


@glycan_network_tools.command("add-prebuilt-neighborhoods", short_help=(
    "Add a set of neighborhoods with built-in rules"))
@click.pass_context
@click.option("-i", "--input-path", type=click.Path(dir_okay=False), default=None)
@click.option("-o", "--output-path", type=click.Path(dir_okay=False, writable=True),
              default=None)
@click.option("-n", "--name", help='Set the neighborhood name', type=click.Choice(
    ["n-glycan", "mammalian-n-glycan", ]), required=True)
def add_prebuild_neighborhoods_to_network(context, input_path, output_path, name):
    if input_path is None:
        input_stream = ctxstream(sys.stdin)
    else:
        input_stream = open(input_path, 'r')
    with input_stream:
        graph = GraphReader(input_stream).network
    if name == 'n-glycan':
        neighborhoods = make_n_glycan_neighborhoods()
    elif name == 'mammalian-n-glycan':
        neighborhoods = make_mammalian_n_glycan_neighborhoods()
    else:
        raise LookupError(name)
    for n in neighborhoods:
        graph.neighborhoods.add(n)
    if output_path is None:
        output_stream = ctxstream(sys.stdout)
    else:
        output_stream = open(output_path, 'w')
    with output_stream:
        GraphWriter(graph, output_stream)


@glycan_network_tools.command("add-neighborhood", short_help=(
    'Add a neighborhood definition to a glycan network'))
@click.pass_context
@click.option("-i", "--input-path", type=click.Path(dir_okay=False), default=None)
@click.option("-o", "--output-path", type=click.Path(dir_okay=False, writable=True),
              default=None)
@click.option("-n", "--name", help='Set the neighborhood name', required=True)
@click.option("-r", "--range-rule", nargs=4, multiple=True, help=(
    "Format: <expression:str> <low:int> <high:int> <required:bool>"
))
@click.option("-e", "--expression-rule", nargs=2, multiple=True, help=(
    "Format: <bool-expression:str> <required:bool>"
))
@click.option("-a", "--ratio-rule", nargs=4, multiple=True, help=(
    "Format: <numerator-expr:str> <denominator-expr:str> <threshold:float> <required:bool>"
))
def add_neighborhood_to_network(context, input_path, output_path, name, range_rule, expression_rule, ratio_rule):
    if input_path is None:
        input_stream = ctxstream(sys.stdin)
    else:
        input_stream = open(input_path, 'r')
    with input_stream:
        graph = GraphReader(input_stream).network
    if name in graph.neighborhoods:
        click.secho(
            "This network already has a neighborhood called %s, overwriting" % name,
            fg='yellow', err=True)
    rules = []
    for rule in range_rule:
        expr, lo, hi, req = rule
        lo = int(lo)
        hi = int(hi)
        req = req.lower().strip() in ('true', 'yes', '1')
        rules.append(CompositionRangeRule(expr, lo, hi, req))
    for rule in expression_rule:
        expr, req = rule
        req = req.lower().strip() in ('true', 'yes', '1')
        rules.append(CompositionExpressionRule(expr, req))
    for rule in ratio_rule:
        numer, denom, threshold, req = rule
        threshold = float(threshold)
        req = req.lower().strip() in ('true', 'yes', '1')
        rules.append(CompositionRatioRule(numer, denom, threshold, req))
    graph.neighborhoods.add(CompositionRuleClassifier(name, rules))
    if output_path is None:
        output_stream = ctxstream(sys.stdout)
    else:
        output_stream = open(output_path, 'w')
    with output_stream:
        GraphWriter(graph, output_stream)


@glycan_network_tools.command("remove-neighborhood", short_help=(
    'Remove a neighborhood definition to a glycan network'))
@click.pass_context
@click.option("-i", "--input-path", type=click.Path(dir_okay=False), default=None)
@click.option("-o", "--output-path", type=click.Path(dir_okay=False, writable=True),
              default=None)
@click.option("-n", "--name", help='Set the neighborhood name', required=True)
def remove_neighborhood(context, input_path, output_path, name):
    if input_path is None:
        input_stream = ctxstream(sys.stdin)
    else:
        input_stream = open(input_path, 'r')
    with input_stream:
        graph = GraphReader(input_stream).network
    try:
        graph.neighborhoods.remove(name)
    except KeyError:
        click.secho(
            "No neighborhood with name %r was found" % name, err=True, fg='yellow')
    if output_path is None:
        output_stream = ctxstream(sys.stdout)
    else:
        output_stream = open(output_path, 'w')
    with output_stream:
        GraphWriter(graph, output_stream)


if __name__ == '__main__':
    cli.main()



cimport cython
from cpython cimport (
    PyTuple_GetItem, PyTuple_Size, PyTuple_GET_ITEM, PyTuple_GET_SIZE,
    PyList_GET_ITEM, PyList_GET_SIZE,
    PySet_Add, PySet_Contains,
    PyDict_GetItem, PyDict_SetItem, PyObject,
    PyFloat_FromDouble, PyFloat_AsDouble,
    PyInt_AsLong)

from libc.math cimport log10, log, sqrt, exp

import numpy as np
cimport numpy as np
np.import_array()
from numpy.math cimport isnan, log2l

from ms_deisotope._c.peak_set cimport DeconvolutedPeak, DeconvolutedPeakSet

from glycresoft._c.structure.fragment_match_map cimport (
    FragmentMatchMap, PeakFragmentPair)

from glycresoft._c.tandem.spectrum_match cimport PeakLabelMap
from glycresoft._c.tandem.oxonium_ions cimport OxoniumIndexMatch

from glypy.composition.ccomposition cimport CComposition

from glycopeptidepy._c.structure.sequence_methods cimport _PeptideSequenceCore
from glycopeptidepy._c.structure.glycan cimport GlycosylationManager
from glycopeptidepy._c.structure.fragment cimport PeptideFragment, FragmentBase, IonSeriesBase, ChemicalShiftBase, SimpleFragment, StubFragment

from glycopeptidepy.structure.fragment import IonSeries, ChemicalShift
from glycopeptidepy.structure.fragmentation_strategy import HCDFragmentationStrategy

cdef:
    IonSeriesBase IonSeries_b, IonSeries_y, IonSeries_c, IonSeries_z, IonSeries_stub_glycopeptide

IonSeries_b = IonSeries.b
IonSeries_y = IonSeries.y
IonSeries_c = IonSeries.c
IonSeries_z = IonSeries.z
IonSeries_stub_glycopeptide = IonSeries.stub_glycopeptide
IonSeries_precursor = IonSeries.precursor


cdef object zeros = np.zeros
cdef object np_float64 = np.float64

@cython.binding(True)
@cython.boundscheck(False)
cpdef _compute_coverage_vectors(self):
    cdef:
        np.ndarray[np.float64_t, ndim=1] n_term_ions, c_term_ions
        int stub_count
        long size
        set glycosylated_n_term_ions, glycosylated_c_term_ions
        IonSeriesBase series
        FragmentMatchMap solution_map
        FragmentBase frag
        PeptideFragment pep_frag
        _PeptideSequenceCore target

    target = <_PeptideSequenceCore>self.target
    size = target.get_size()

    n_term_ions = zeros(size, dtype=np_float64)
    c_term_ions = zeros(size, dtype=np_float64)
    stub_count = 0
    glycosylated_n_term_ions = set()
    glycosylated_c_term_ions = set()

    solution_map = self.solution_map

    for obj in solution_map.fragments():
        frag = <FragmentBase>obj
        series = frag.get_series()
        if series.int_code == IonSeries_b.int_code or series.int_code == IonSeries_c.int_code:
            pep_frag = <PeptideFragment>frag
            n_term_ions[pep_frag.position] = 1
            if pep_frag._is_glycosylated():
                glycosylated_n_term_ions.add((series, pep_frag.position))
        elif series.int_code == IonSeries_y.int_code or series.int_code == IonSeries_z.int_code:
            pep_frag = <PeptideFragment>frag
            c_term_ions[pep_frag.position] = 1
            if pep_frag._is_glycosylated():
                glycosylated_c_term_ions.add((series, pep_frag.position))
        elif series.int_code == IonSeries_stub_glycopeptide.int_code:
            stub_count += 1

    # Ignore b1 ions. They are indistinguishable from immonium ions and
    # are not meaningful
    if size > 0:
        n_term_ions[0] = 0
    return n_term_ions, c_term_ions, stub_count, len(glycosylated_n_term_ions), len(glycosylated_c_term_ions)


@cython.binding(True)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef double _calculate_peptide_coverage_no_glycosylated(self):
    cdef:
        np.ndarray[np.float64_t, ndim=1] n_term_ions, c_term_ions
        int stub_count
        long size
        double coverage_score
        size_t i
        set glycosylated_n_term_ions, glycosylated_c_term_ions
        IonSeriesBase series
        FragmentMatchMap solution_map
        FragmentBase frag
        PeptideFragment pep_frag
        _PeptideSequenceCore target
        PeakFragmentPair pfp

    target = <_PeptideSequenceCore>self.target
    solution_map = self.solution_map
    size = target.get_size()

    n_term_ions = zeros(size, dtype=np_float64)
    c_term_ions = zeros(size, dtype=np_float64)

    for obj in solution_map.fragments():
        frag = <FragmentBase>obj
        series = frag.get_series()
        if series.int_code == IonSeries_b.int_code or series.int_code == IonSeries_c.int_code:
            pep_frag = <PeptideFragment>frag
            if not pep_frag._is_glycosylated():
                n_term_ions[pep_frag.position] = 1.0
        if series.int_code == IonSeries_y.int_code or series.int_code == IonSeries_z.int_code:
            pep_frag = <PeptideFragment>frag
            if not pep_frag._is_glycosylated():
                c_term_ions[pep_frag.position] = 1.0

    # Ignore b1 ions. They are indistinguishable from immonium ions and
    # are not meaningful
    if size > 0:
        n_term_ions[0] = 0
    coverage_score = 0.0
    for i in range(size):
        coverage_score += (n_term_ions[i] or c_term_ions[size - i - 1])
    return coverage_score / size


cdef set peptide_ion_series = {IonSeries_b, IonSeries_y, IonSeries_c, IonSeries_z}
cdef int PEPTIDE_SERIES_NUMBER = IonSeries_z.int_code + 1


@cython.binding(True)
@cython.cdivision(True)
@cython.boundscheck(False)
def calculate_peptide_score(self, double error_tolerance=2e-5, double peptide_coverage_weight=1.0, *args, **kwargs):
    cdef:
        double total, score, coverage_score
        PeakFragmentPair peak_pair
        DeconvolutedPeak peak
        FragmentMatchMap solution_map

    target = <_PeptideSequenceCore>self.target
    size = target.get_size()

    total = 0
    solution_map = <FragmentMatchMap>self.solution_map
    for obj in solution_map.members:
        peak_pair = <PeakFragmentPair>obj
        peak = peak_pair.peak
        if (<FragmentBase>peak_pair.fragment).get_series().int_code < PEPTIDE_SERIES_NUMBER:
            total += log10(peak.intensity) * (1 - (abs(peak_pair.mass_accuracy()) / error_tolerance) ** 4)

    coverage_score = _calculate_peptide_coverage(self)

    score = total * coverage_score ** peptide_coverage_weight
    if isnan(score):
        return 0
    return score


@cython.binding(True)
@cython.cdivision(True)
@cython.boundscheck(False)
def calculate_peptide_score_no_glycosylated(self, double error_tolerance=2e-5, double peptide_coverage_weight=0.7, *args, **kwargs):
    cdef:
        double total, score, coverage_score
        PeakFragmentPair peak_pair
        DeconvolutedPeak peak
        FragmentMatchMap solution_map
        bint is_exd

    is_exd = self.is_exd()
    target = <_PeptideSequenceCore>self.target
    size = target.get_size()

    total = 0
    solution_map = <FragmentMatchMap>self.solution_map
    for obj in solution_map.members:
        peak_pair = <PeakFragmentPair>obj
        peak = peak_pair.peak
        if (<FragmentBase>peak_pair.fragment).get_series().int_code < PEPTIDE_SERIES_NUMBER:
            if (not (<PeptideFragment>peak_pair.fragment)._is_glycosylated()) or is_exd:
                total += log10(peak.intensity) * (1 - (abs(peak_pair.mass_accuracy()) / error_tolerance) ** 4)

    coverage_score = _calculate_peptide_coverage_no_glycosylated(self)

    score = total * coverage_score ** peptide_coverage_weight
    if isnan(score):
        return 0
    return score


@cython.binding(True)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef double _calculate_peptide_coverage(self):
    cdef:
        double coverage_score, normalizer
        tuple coverage_result
        long size
        # np.ndarray[np.float64_t, ndim=1, mode='c'] n_term, c_term
        np.float64_t[::1] n_term, c_term
        size_t i

    cache = self._peptide_coverage
    if cache is not None:
        return cache

    target = <_PeptideSequenceCore>self.target
    size = target.get_size()

    coverage_result = <tuple>_compute_coverage_vectors(self)
    n_term = <np.ndarray[np.float64_t, ndim=1, mode='c']>PyTuple_GET_ITEM(coverage_result, 0)
    c_term = <np.ndarray[np.float64_t, ndim=1, mode='c']>PyTuple_GET_ITEM(coverage_result, 1)
    normalizer = float((2 * size - 1))
    coverage_score = 0.0

    # Ignore b1 ions. They are indistinguishable from immonium ions and
    # are not meaningful
    if size > 0:
        n_term[0] = 0
    for i in range(size):
        coverage_score += n_term[i] + c_term[size - i - 1]
    coverage_score /= normalizer
    self._peptide_coverage = coverage_score
    return coverage_score


@cython.binding(True)
@cython.cdivision(True)
@cython.boundscheck(False)
def calculate_glycan_score(self, double error_tolerance=2e-5, double glycan_core_weight=0.4, double glycan_coverage_weight=0.5,
                           bint fragile_fucose=False, bint extended_glycan_search=False, *args, **kwargs):
    cdef:
        set seen, core_fragments, core_matches, extended_matches
        IonSeriesBase series
        list theoretical_set
        double total, score
        double glycan_prior, glycan_coverage
        FragmentMatchMap solution_map
        FragmentBase frag
        PeakFragmentPair peak_pair
        DeconvolutedPeak peak
        int side_group_count
        size_t i, m
        object target
    target = self.target
    seen = set()
    series = IonSeries_stub_glycopeptide
    if not extended_glycan_search:
        theoretical_set = list(target.stub_fragments(extended=True))
    else:
        theoretical_set = list(target.stub_fragments(extended=True, extended_fucosylation=True))
    core_fragments = set()
    for i in range(len(theoretical_set)):
        frag = <FragmentBase>theoretical_set[i]
        if not frag.is_extended:
            core_fragments.add(frag._name)

    total = 0
    core_matches = set()
    extended_matches = set()
    solution_map = <FragmentMatchMap>self.solution_map

    for obj in solution_map.members:
        peak_pair = <PeakFragmentPair>obj
        if (<FragmentBase>peak_pair.fragment).get_series().int_code != series.int_code:
            continue
        fragment_name = (<FragmentBase>peak_pair.fragment).base_name()
        if fragment_name in core_fragments:
            core_matches.add(fragment_name)
        else:
            extended_matches.add(fragment_name)
        peak = peak_pair.peak
        if peak._index.neutral_mass not in seen:
            seen.add(peak._index.neutral_mass)
            total += log10(peak.intensity) * (1 - (abs(peak_pair.mass_accuracy()) / error_tolerance) ** 4)

    m = PyList_GET_SIZE(theoretical_set)
    glycan_coverage = _compute_glycan_coverage_from_metrics(
        self, fragile_fucose, len(core_matches), len(extended_matches),
        len(core_fragments), m, glycan_core_weight, glycan_coverage_weight)
    score = total * glycan_coverage
    glycan_prior = 0.0
    self._glycan_coverage = glycan_coverage
    if glycan_coverage > 0:
        glycan_prior = target.glycan_prior
        score += glycan_coverage * glycan_prior
    if isnan(score):
        return 0
    return score


@cython.binding(True)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef double _calculate_glycan_coverage(self, double core_weight=0.4, double coverage_weight=0.5,
                                        bint fragile_fucose=False, bint extended_glycan_search=False):
    cdef:
        set seen, core_fragments, core_matches, extended_matches
        IonSeriesBase series
        list theoretical_set
        double total, n, k, d, core_coverage, extended_coverage, score
        FragmentMatchMap solution_map
        StubFragment frag
        PeakFragmentPair peak_pair
        size_t i, n_core_matches, n_extended_matches, m
        str frag_name

    if self._glycan_coverage is not None:
            return self._glycan_coverage
    seen = set()
    series = IonSeries_stub_glycopeptide
    if not extended_glycan_search:
        theoretical_set = list(self.target.stub_fragments(extended=True))
    else:
        theoretical_set = list(self.target.stub_fragments(extended=True, extended_fucosylation=True))
    core_fragments = set()
    m = len(theoretical_set)
    for i in range(m):
        frag = <StubFragment>theoretical_set[i]
        if not frag.is_extended:
            core_fragments.add(frag._name)

    total = 0
    core_matches = set()
    extended_matches = set()
    solution_map = <FragmentMatchMap>self.solution_map

    for obj in solution_map.members:
        peak_pair = <PeakFragmentPair>obj
        if (<FragmentBase>peak_pair.fragment).get_series().int_code != series.int_code:
            continue
        frag_name = peak_pair.fragment.base_name()
        if frag_name in core_fragments:
            core_matches.add(frag_name)
        else:
            extended_matches.add(frag_name)

    coverage = self._compute_glycan_coverage_from_metrics(
        fragile_fucose, len(core_matches), len(extended_matches),
        len(core_fragments), m, core_weight, coverage_weight)

    self._glycan_coverage = coverage
    return coverage


@cython.binding(True)
@cython.cdivision(True)
cpdef double _compute_glycan_coverage_from_metrics(self, bint fragile_fucose, size_t n_core_matches,
                                                   size_t n_extended_matches, size_t n_core_fragments, size_t n_fragments,
                                                   double core_weight, double coverage_weight):
    cdef:
        double glycan_composition_size_normalizer
        double core_coverage, extended_coverage
        double approximate_size, extra_branch_factor

        double coverage


    glycan_composition = self.target.glycan_composition
    approximate_size = self._get_internal_size(glycan_composition)
    extra_branch_factor = 2.0
    if not fragile_fucose:
        side_group_count = self._glycan_side_group_count(glycan_composition)
        if side_group_count > 0:
            extra_branch_factor = 1.0

    glycan_composition_size_normalizer = min(
        max(approximate_size * log(approximate_size) / extra_branch_factor, approximate_size),
        n_fragments)

    core_coverage = ((n_core_matches * 1.0) / n_core_fragments) ** core_weight
    extended_coverage = min((n_core_matches + n_extended_matches) / glycan_composition_size_normalizer, 1.0) ** coverage_weight
    coverage = core_coverage * extended_coverage
    if isnan(coverage):
        coverage = 0.0
    return coverage

@cython.binding(True)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef SimpleCoverageScorer_match_backbone_series(self, IonSeriesBase series, double error_tolerance=2e-5,
                                                 set masked_peaks=None, strategy=None, bint include_neutral_losses=False):
    cdef:
        list frags, fragments
        tuple peaks
        bint glycosylated_position, previous_position_glycosylated
        PeptideFragment frag
        FragmentMatchMap solution_map
        long glycosylated_term_ions_count
        DeconvolutedPeak peak
        DeconvolutedPeakSet spectrum
        size_t i, n, j, m, i_peaks, n_peaks

    if strategy is None:
        strategy = HCDFragmentationStrategy
    # Assumes that fragmentation proceeds from the start of the ladder (series position 1)
    # which means that if the last fragment could be glycosylated then the next one will be
    # but if the last fragment wasn't the next one might be.
    previous_position_glycosylated = False
    glycosylated_term_ions_count = 0

    spectrum = self.spectrum
    obj = self.get_fragments(series, strategy=strategy, include_neutral_losses=include_neutral_losses)
    if not isinstance(obj, list):
        fragments = list(obj)
    else:
        fragments = <list>obj

    solution_map = <FragmentMatchMap>self.solution_map

    n = PyList_GET_SIZE(fragments)
    for i in range(n):
        frags = <list>PyList_GET_ITEM(fragments, i)
        glycosylated_position = previous_position_glycosylated
        m = PyList_GET_SIZE(frags)
        for j in range(m):
            frag = <PeptideFragment>PyList_GET_ITEM(frags, j)
            if not glycosylated_position:
                glycosylated_position |= frag._is_glycosylated()
            peaks = <tuple>spectrum.all_peaks_for(frag.mass, error_tolerance)
            for i_peaks in range(PyTuple_GET_SIZE(peaks)):
                peak = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks, i_peaks)
                if peak._index.neutral_mass in masked_peaks:
                    continue
                FragmentMatchMap._add_direct(solution_map, PeakFragmentPair._create_simple(peak, frag))
        if glycosylated_position:
            glycosylated_term_ions_count += 1
        previous_position_glycosylated = glycosylated_position
    if series.direction > 0:
        self.glycosylated_n_term_ion_count += glycosylated_term_ions_count
    else:
        self.glycosylated_c_term_ion_count += glycosylated_term_ions_count



@cython.binding(True)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef CoverageWeightedBinomialScorer_match_backbone_series(self, IonSeriesBase series, double error_tolerance=2e-5,
                                                           set masked_peaks=None, strategy=None,
                                                           bint include_neutral_losses=False):
    cdef:
        list frags, fragments
        tuple peaks
        bint glycosylated_position, previous_position_glycosylated
        PeptideFragment frag
        FragmentMatchMap solution_map
        long glycosylated_term_ions_count
        long n_theoretical
        DeconvolutedPeak peak
        DeconvolutedPeakSet spectrum
        size_t i, n, j, m, i_peaks, n_peaks


    if strategy is None:
        strategy = HCDFragmentationStrategy

    n_theoretical = 0
    previous_position_glycosylated = False
    glycosylated_term_ions_count = 0

    spectrum = self.spectrum
    obj = self.get_fragments(series, strategy=strategy, include_neutral_losses=include_neutral_losses)
    if not isinstance(obj, list):
        fragments = list(obj)
    else:
        fragments = <list>obj

    solution_map = <FragmentMatchMap>self.solution_map

    n = PyList_GET_SIZE(fragments)
    for i in range(n):
        frags = <list>PyList_GET_ITEM(fragments, i)

        glycosylated_position = previous_position_glycosylated
        n_theoretical += 1
        m = PyList_GET_SIZE(frags)
        for j in range(m):
            frag = <PeptideFragment>PyList_GET_ITEM(frags, j)
            if not glycosylated_position:
                glycosylated_position |= frag._is_glycosylated()
            peaks = spectrum.all_peaks_for(frag.mass, error_tolerance)
            n_peaks = PyTuple_GET_SIZE(peaks)
            for i_peaks in range(n_peaks):
                peak = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks, i_peaks)
                if peak._index.neutral_mass in masked_peaks:
                    continue
                solution_map.add(peak, frag)
        if glycosylated_position:
            glycosylated_term_ions_count += 1
        previous_position_glycosylated = glycosylated_position

    self.n_theoretical += n_theoretical
    if series.direction > 0:
        self.glycosylated_n_term_ion_count += glycosylated_term_ions_count
    else:
        self.glycosylated_c_term_ion_count += glycosylated_term_ions_count


cdef set decode_oxonium_index_match(_PeptideSequenceCore target, OxoniumIndexMatch index_match, FragmentMatchMap solution_map, set masked_peaks):
    cdef:
        size_t i, j, m, n
        Py_ssize_t last
        DeconvolutedPeak peak
        SimpleFragment frag
        PeakFragmentPair pfp
        list fragment_index_pairs
        object frag_index_pair, k, gc

    gc = (<GlycosylationManager>target._glycosylation_manager).get_glycan_composition()
    fragment_index_pairs = index_match.by_glycan(str(gc))
    if fragment_index_pairs is None:
        return None

    n = PyList_GET_SIZE(fragment_index_pairs)

    for i in range(n):
        pfp = <PeakFragmentPair>PyList_GET_ITEM(fragment_index_pairs, i)
        k = pfp.peak._index.neutral_mass
        # frag_index_pair = <object>PyList_GET_ITEM(fragment_index_pairs, i)
        # frag = <SimpleFragment>PyTuple_GET_ITEM(frag_index_pair, 0)
        # k = <object>PyTuple_GET_ITEM(frag_index_pair, 1)
        if PySet_Contains(masked_peaks, k):
            continue
        PySet_Add(masked_peaks, k)
        # j = PyInt_AsLong(k)
        # peak = spectrum.getitem(j)
        # solution_map.add(peak, frag)
        FragmentMatchMap._add_direct(solution_map, pfp)
    return masked_peaks


@cython.binding(True)
@cython.nonecheck(False)
cpdef _match_oxonium_ions(self, double error_tolerance=2e-5, set masked_peaks=None):
    cdef:
        list fragments
        FragmentBase frag
        DeconvolutedPeak peak
        DeconvolutedPeakSet spectrum
        FragmentMatchMap solution_map
        PeakLabelMap label_map
        _PeptideSequenceCore target
        object ix
        dict scan_annotations
        PyObject* tmp
        bint checked
        OxoniumIndexMatch index_match

    if masked_peaks is None:
        masked_peaks = set()

    result = None
    target = <_PeptideSequenceCore>self.target
    spectrum = <DeconvolutedPeakSet>self.spectrum
    solution_map = <FragmentMatchMap>self.solution_map
    scan_annotations = <dict>self.scan.annotations
    tmp = PyDict_GetItem(scan_annotations, 'oxonium_index_match')
    if tmp != NULL:
        index_match = <OxoniumIndexMatch>tmp
        result = decode_oxonium_index_match(target, index_match, solution_map, masked_peaks)
        if result is not None:
            return masked_peaks

    tmp = PyDict_GetItem(scan_annotations, 'peak_label_map')
    if tmp == NULL:
        peak_label_map = PeakLabelMap._create()
        PyDict_SetItem(scan_annotations, 'peak_label_map', peak_label_map)
    else:
        peak_label_map = <PeakLabelMap>tmp

    obj = self.target.glycan_fragments()

    if isinstance(obj, list):
        fragments = <list>obj
    else:
        fragments = list(obj)

    checked = False
    # TODO: Instead of looping over all the fragments all the time,
    # instead use a map from glycan compositions to list of (fragment, peak)
    # pairs? Really only useful if the thing dominating the runtime isn't the
    # solution_map.add() call.
    for i in range(len(fragments)):
        frag = <FragmentBase>PyList_GET_ITEM(fragments, i)
        peak = peak_label_map.get(frag._name, &checked)
        if not checked:
            peak = spectrum.has_peak(frag.mass, error_tolerance)
            peak_label_map.add(frag._name, peak)

        if peak is not None:
            ix = peak._index.neutral_mass
            if ix not in masked_peaks:
                solution_map.add(peak, frag)
                masked_peaks.add(ix)
    return masked_peaks


cdef double INF = float("inf")


@cython.binding(True)
cpdef _match_stub_glycopeptides(self, double error_tolerance=2e-5, set masked_peaks=None,
                                ChemicalShiftBase chemical_shift=None, bint extended_glycan_search=False):

    cdef:
        list fragments
        tuple peaks, shifted_peaks
        DeconvolutedPeak peak
        size_t i, n, j, m, k
        DeconvolutedPeakSet spectrum
        FragmentMatchMap solution_map
        StubFragment frag, shifted_frag
        double max_mass, current_mass
        PeakFragmentPair pfp

    if masked_peaks is None:
        masked_peaks = set()

    if not extended_glycan_search:
        obj = self.target.stub_fragments(extended=True)
    else:
        obj = self.target.stub_fragments(extended=True, extended_fucosylation=True)
    if isinstance(obj, list):
        fragments = <list>obj
    else:
        fragments = list(obj)

    spectrum = <DeconvolutedPeakSet>self.spectrum
    solution_map = <FragmentMatchMap>self.solution_map

    max_mass = INF
    n = spectrum.get_size()
    if n > 0:
        max_mass = spectrum.getitem(n - 1).neutral_mass + 2
    else:
        max_mass = 0.0

    n = PyList_GET_SIZE(fragments)
    for i in range(n):
        frag = <StubFragment>PyList_GET_ITEM(fragments, i)
        current_mass = frag.mass

        if current_mass > max_mass:
            continue

        peaks = spectrum.all_peaks_for(current_mass, error_tolerance)
        for j in range(PyTuple_Size(peaks)):
            peak = <DeconvolutedPeak>PyTuple_GetItem(peaks, j)
            # should we be masking these? peptides which have amino acids which are
            # approximately the same mass as a monosaccharide unit at ther terminus
            # can produce cases where a stub ion and a backbone fragment match the
            # same peak.
            #
            PySet_Add(masked_peaks, peak._index.neutral_mass)
            pfp = PeakFragmentPair._create_simple(peak, frag)
            FragmentMatchMap._add_direct(solution_map, pfp)
        if chemical_shift is not None:
            shifted_mass = frag.mass + chemical_shift.mass
            if shifted_mass > max_mass:
                continue
            shifted_peaks = spectrum.all_peaks_for(shifted_mass, error_tolerance)
            for k in range(PyTuple_Size(shifted_peaks)):
                peak = <DeconvolutedPeak>PyTuple_GetItem(shifted_peaks, k)
                PySet_Add(masked_peaks, peak._index.neutral_mass)

                shifted_frag = frag.clone()
                shifted_frag.set_chemical_shift(chemical_shift)
                pfp = PeakFragmentPair._create_simple(peak, shifted_frag)
                FragmentMatchMap._add_direct(solution_map, pfp)
    return masked_peaks


@cython.binding(True)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef base_peak(self):
    cdef:
        DeconvolutedPeak peak
        DeconvolutedPeakSet spectrum
        size_t i, n
        dict annotations
        double max_intensity
        PyObject* tmp
    scan = self.scan
    annotations = <dict>scan.annotations
    tmp = PyDict_GetItem(annotations, "_base_peak_intensity")
    if tmp != NULL:
        return <object>tmp
    spectrum = <DeconvolutedPeakSet>self.spectrum
    if spectrum is None:
        return 0
    n = spectrum.get_size()
    max_peak = None
    max_intensity = 0
    for i in range(n):
        peak = spectrum.getitem(i)
        if peak.intensity > max_intensity:
            max_intensity = peak.intensity
            max_peak = peak
    PyDict_SetItem(annotations, "_base_peak_intensity", PyFloat_FromDouble(max_intensity))
    return max_intensity


@cython.boundscheck(False)
cpdef DeconvolutedPeak base_peak_tuple(tuple peaks):
    cdef:
        size_t i, n
        DeconvolutedPeak peak, best_peak

    n = PyTuple_Size(peaks)
    if n == 0:
        return None
    else:
        best_peak = <DeconvolutedPeak>PyTuple_GetItem(peaks, 0)
    for i in range(1, n):
        peak = <DeconvolutedPeak>PyTuple_GetItem(peaks, i)
        if peak.intensity > best_peak.intensity:
            best_peak = peak
    return best_peak


cdef double sqrt2pi = sqrt(2 * np.pi)


ctypedef fused scalar_or_array:
    double
    np.ndarray


@cython.cdivision(True)
cpdef scalar_or_array gauss(scalar_or_array x, double mu, double sigma):
    if scalar_or_array is np.ndarray:
        return np.exp(-((x - mu) ** 2) / (2 * sigma ** 2)) / (sigma * sqrt2pi)
    else:
        return exp(-((x - mu) ** 2) / (2 * sigma ** 2)) / (sigma * sqrt2pi)


@cython.binding(True)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef PeptideSpectrumMatcherBase_match_backbone_series(self, IonSeriesBase series, double error_tolerance=2e-5,
                                                       set masked_peaks=None, strategy=None, bint include_neutral_losses=False):
    cdef:
        list frags, fragments
        tuple peaks
        PeptideFragment frag
        FragmentMatchMap solution_map
        DeconvolutedPeak peak
        DeconvolutedPeakSet spectrum
        size_t i, n, j, m, i_peaks, n_peaks

    if strategy is None:
        strategy = HCDFragmentationStrategy

    spectrum = self.spectrum
    obj = self.get_fragments(series, strategy=strategy, include_neutral_losses=include_neutral_losses)
    if not isinstance(obj, list):
        fragments = list(obj)
    else:
        fragments = <list>obj

    solution_map = <FragmentMatchMap>self.solution_map

    n = PyList_GET_SIZE(fragments)
    for i in range(n):
        frags = <list>PyList_GET_ITEM(fragments, i)
        m = PyList_GET_SIZE(frags)
        for j in range(m):
            frag = <PeptideFragment>PyList_GET_ITEM(frags, j)
            peaks = <tuple>spectrum.all_peaks_for(frag.mass, error_tolerance)
            n_peaks = PyTuple_GET_SIZE(peaks)
            for i_peaks in range(n_peaks):
                peak = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks, i_peaks)
                if PySet_Contains(masked_peaks, peak._index.neutral_mass):
                    continue
                solution_map.add(peak, frag)


@cython.binding(True)
@cython.boundscheck(False)
cpdef tuple _peptide_compute_coverage_vectors(self):
    cdef:
        np.ndarray[np.float64_t, ndim=1] n_term_ions, c_term_ions
        long size
        FragmentMatchMap solution_map
        FragmentBase frag
        PeptideFragment pep_frag
        _PeptideSequenceCore target

    target = <_PeptideSequenceCore>self.target
    size = target.get_size()

    n_term_ions = zeros(size, dtype=np_float64)
    c_term_ions = zeros(size, dtype=np_float64)

    solution_map = self.solution_map

    for obj in solution_map.fragments():
        frag = <FragmentBase>obj
        series = frag.get_series()
        if series.name in (IonSeries_b.name, IonSeries_c.name):
            pep_frag = <PeptideFragment>frag
            n_term_ions[pep_frag.position] = 1
        elif series.name in (IonSeries_y.name, IonSeries_z.name):
            pep_frag = <PeptideFragment>frag
            c_term_ions[pep_frag.position] = 1
    return n_term_ions, c_term_ions


cdef double log2_3 = log2l(3)

@cython.binding(True)
@cython.boundscheck(False)
@cython.cdivision(True)
cpdef compute_coverage(self):
    '''Computes log-weighted coverage over the peptide sequence.

    The log-weighting scheme rewards more for the first fragment than
    the second fragment which may be advantageous for some scenarios.
    '''
    cdef:
        np.ndarray[np.float64_t, ndim=1, mode='c'] n_term_ions, c_term_ions
        long size, i
        _PeptideSequenceCore target
        double acc, mean_coverage
        tuple vectors

    target = <_PeptideSequenceCore>self.target
    size = target.get_size()
    vectors = <tuple>self._compute_coverage_vectors()
    n_term_ions = <np.ndarray>PyTuple_GetItem(vectors, 0)
    c_term_ions = <np.ndarray>PyTuple_GetItem(vectors, 1)
    acc = 0.0

    for i in range(size):
        acc += log2l(n_term_ions[i] + c_term_ions[size - (i + 1)] + 1) / log2_3
    mean_coverage = acc / size
    return mean_coverage


@cython.binding(True)
cpdef count_peptide_Y_ion_utilization(self):
    cdef:
        float weight
        int stub_count, peptide_backbone_count
        int series_code
        object cached_value, tmp
        PeakFragmentPair pfp
        FragmentMatchMap solution_map
        FragmentBase frag

    cached_value = self._peptide_Y_ion_utilization
    if cached_value is not None:
        return cached_value
    solution_map = <FragmentMatchMap>self.solution_map
    weight = 0.0
    stub_count = 0
    peptide_backbone_count = 0
    for tmp in solution_map.members:
        pfp = <PeakFragmentPair>tmp
        frag = <FragmentBase>pfp.fragment
        series_code = frag.get_series().int_code
        if series_code == IonSeries_stub_glycopeptide.int_code:
            stub_count += 1
            weight += log10(pfp.peak.intensity)
        elif series_code <= IonSeries_z.int_code:
            peptide_backbone_count += 1
    self._peptide_Y_ion_utilization = (weight, stub_count, peptide_backbone_count)
    return self._peptide_Y_ion_utilization


cpdef parse_float(str text):
    cdef:
        object value
    value = float(text)
    if isnan(value):
        return 0.0
    return value


cdef CComposition H2O_loss = -CComposition("H2O")
cdef CComposition NH3_loss = -CComposition("NH3")

cdef double NH3_loss_mass = NH3_loss.mass
cdef double H2O_loss_mass = H2O_loss.mass

cdef ChemicalShiftBase NH3_loss_shift = ChemicalShift("-NH3", NH3_loss)
cdef ChemicalShiftBase H2O_loss_shift = ChemicalShift("-H2O", H2O_loss)

@cython.binding(True)
cpdef _match_precursor(self, double error_tolerance=2e-5, set masked_peaks=None, bint include_neutral_losses=False):

    cdef:
        list frags, fragments
        tuple peaks
        SimpleFragment frag, shifted_frag
        FragmentMatchMap solution_map
        DeconvolutedPeak peak
        DeconvolutedPeakSet spectrum
        size_t i_peaks

    if masked_peaks is None:
        masked_peaks = set()

    mass = self.target.total_mass
    frag = SimpleFragment("M", mass, IonSeries_precursor, None)

    spectrum = self.spectrum
    solution_map = <FragmentMatchMap>self.solution_map
    peaks = <tuple>spectrum.all_peaks_for(frag.mass, error_tolerance)
    for i_peaks in range(PyTuple_Size(peaks)):
        peak = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks, i_peaks)
        key = <object>peak._index.neutral_mass
        if PySet_Contains(masked_peaks, key):
            continue
        PySet_Add(masked_peaks, key)
        solution_map.add(peak, frag)
    if include_neutral_losses:
        peaks = <tuple>spectrum.all_peaks_for(frag.mass + NH3_loss_mass, error_tolerance)
        for i_peaks in range(PyTuple_Size(peaks)):
            peak = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks, i_peaks)
            key = <object>peak._index.neutral_mass
            if PySet_Contains(masked_peaks, key):
                continue
            PySet_Add(masked_peaks, key)
            shifted_frag = frag.clone()
            shifted_frag.set_chemical_shift(NH3_loss_shift)
            solution_map.add(peak, shifted_frag)
        peaks = <tuple>spectrum.all_peaks_for(frag.mass + H2O_loss_mass, error_tolerance)
        for i_peaks in range(PyTuple_Size(peaks)):
            peak = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks, i_peaks)
            key = <object>peak._index.neutral_mass
            if PySet_Contains(masked_peaks, key):
                continue
            PySet_Add(masked_peaks, key)
            shifted_frag = frag.clone()
            shifted_frag.set_chemical_shift(H2O_loss_shift)
            solution_map.add(peak, shifted_frag)



cdef long _factorial(long x):
    cdef:
        long i
        long acc

    acc = 1
    for i in range(x, 0, -1):
        acc *= i
    return acc


@cython.binding(True)
def _calculate_hyperscore(self, *args, **kwargs):
    cdef:
        FragmentMatchMap solution_map
        PeakFragmentPair pfp
        PeptideFragment pep_frag
        double n_term_intensity
        double c_term_intensity
        long n_term
        long c_term
        double hyper


    n_term_intensity = 0
    c_term_intensity = 0
    n_term = 0
    c_term = 0
    hyper = 0

    solution_map = self.solution_map

    for obj in solution_map.members:
        pfp = <PeakFragmentPair>obj
        series = (<FragmentBase>pfp.fragment).get_series()
        if series.name in (IonSeries_b.name, IonSeries_c.name):
            n_term += 1
            n_term_intensity += pfp.peak.intensity
        elif series.name in (IonSeries_y.name, IonSeries_z.name):
            c_term += 1
            c_term_intensity += pfp.peak.intensity

    hyper += log(n_term_intensity)
    hyper += log(_factorial(n_term))

    hyper += log(c_term_intensity)
    hyper += log(_factorial(c_term))
    return hyper


cpdef double fast_coarse_score(DeconvolutedPeakSet peak_set, _PeptideSequenceCore target, double error_tolerance=2e-5):
    cdef:
        size_t i, n, j, m, k, l
        list fragments, frags
        double acc
        object fragout
        PeptideFragment frag
        tuple peaks
        DeconvolutedPeak peak

    acc = 0.0
    fragout = target.get_fragments("b")
    if isinstance(fragout, list):
        fragments = <list>fragout
    else:
        fragments = list(fragout)
    n = PyList_GET_SIZE(fragments)
    for i in range(n):
        frags = <list>PyList_GET_ITEM(fragments, i)
        m = PyList_GET_SIZE(frags)
        for j in range(m):
            frag = <PeptideFragment>PyList_GET_ITEM(frags, j)
            peaks = peak_set.all_peaks_for(frag.mass, error_tolerance)
            l = PyTuple_GET_SIZE(peaks)
            for k in range(l):
                peak = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks, k)
                acc += log10(peak.intensity)

    fragout = target.get_fragments("y")
    if isinstance(fragout, list):
        fragments = <list>fragout
    else:
        fragments = list(fragout)
    n = PyList_GET_SIZE(fragments)
    for i in range(n):
        frags = <list>PyList_GET_ITEM(fragments, i)
        m = PyList_GET_SIZE(frags)
        for j in range(m):
            frag = <PeptideFragment>PyList_GET_ITEM(frags, j)
            peaks = peak_set.all_peaks_for(frag.mass, error_tolerance)
            l = PyTuple_GET_SIZE(peaks)
            for k in range(l):
                peak = <DeconvolutedPeak>PyTuple_GET_ITEM(peaks, k)
                acc += log10(peak.intensity)
    return acc


@cython.binding(True)
cpdef correlate_fragment_map(self, other):
    cdef:
        dict self_map, other_map
        double x, y, xx, yy, xy
        void* ptmp

    self_map = self.fragment_map
    other_map = other.fragment_map
    xy = 0
    xx = 0
    yy = 0
    keyspace = set(self_map) | set(other_map)
    for k in keyspace:
        ptmp = PyDict_GetItem(self_map, k)
        if ptmp == NULL:
            x = 0
        else:
            x = PyFloat_AsDouble(<object>ptmp)

        ptmp = PyDict_GetItem(other_map, k)
        if ptmp == NULL:
            y = 0
        else:
            y = PyFloat_AsDouble(<object>ptmp)

        xy += x * y
        xx += x ** 2
        yy += y ** 2
    xx = sqrt(xx)
    yy = sqrt(yy)
    if xx == 0 or yy == 0:
        return 0
    return xy / (xx * yy)




cimport cython


cdef class MassObject(object):
    cdef:
        public object obj
        public double mass


cpdef (ssize_t, bint) binary_search_with_flag(list array, double mass, double error_tolerance=*)


cdef class ChromatogramIndex(object):
    cdef:
        list _chromatograms

    cpdef (ssize_t, bint) _binary_search(self, double mass, double error_tolerance=*)
    cpdef object find_all_by_mass(self, double mass, double ppm_error_tolerance=*)
    cpdef object find_mass(self, double mass, double ppm_error_tolerance=*)
    cpdef object mass_between(self, double low, double high)

    cdef inline MassObject get_index(self, size_t i)
    cdef inline size_t get_size(self)





import math
import numpy as np

from glycopeptidepy.structure.fragment import IonSeries

from .base import ModelTreeNode
from .precursor_mass_accuracy import MassAccuracyMixin
from .simple_score import SignatureAwareCoverageScorer


class LogIntensityScorer(SignatureAwareCoverageScorer, MassAccuracyMixin):

    _peptide_score = None
    _glycan_score = None

    def __init__(self, scan, sequence, mass_shift=None, *args, **kwargs):
        super(LogIntensityScorer, self).__init__(scan, sequence, mass_shift, *args, **kwargs)

    def calculate_score(self, error_tolerance=2e-5, peptide_weight=0.65, *args, **kwargs) -> float:
        glycan_weight = 1 - peptide_weight
        combo_score = self.peptide_score(error_tolerance, *args, **kwargs) * peptide_weight +\
                      self.glycan_score(error_tolerance, *args, **kwargs) * glycan_weight
        mass_accuracy = self._precursor_mass_accuracy_score()
        signature_component = self._signature_ion_score()
        self._score = combo_score + mass_accuracy + signature_component
        return self._score

    def calculate_peptide_score(self, error_tolerance=2e-5, peptide_coverage_weight=1.0, *args, **kwargs) -> float:
        total = 0
        series_set = {IonSeries.b, IonSeries.y, IonSeries.c, IonSeries.z}
        seen = set()
        for peak_pair in self.solution_map:
            peak = peak_pair.peak
            if peak_pair.fragment.get_series() in series_set:
                seen.add(peak.index.neutral_mass)
                total += np.log10(peak.intensity) * (1 - (abs(peak_pair.mass_accuracy()) / error_tolerance) ** 4)
        n_term, c_term = self._compute_coverage_vectors()[:2]
        coverage_score = ((n_term + c_term[::-1])).sum() / float((2 * len(self.target) - 1))
        score = total * coverage_score ** peptide_coverage_weight
        if np.isnan(score):
            return 0
        return score

    def calculate_glycan_score(self, error_tolerance=2e-5, glycan_core_weight=0.4, glycan_coverage_weight=0.5,
                               fragile_fucose=False, extended_glycan_search=False, *args, **kwargs) -> float:
        seen = set()
        series = IonSeries.stub_glycopeptide
        if not extended_glycan_search:
            theoretical_set = list(self.target.stub_fragments(extended=True))
        else:
            theoretical_set = list(self.target.stub_fragments(extended=True, extended_fucosylation=True))
        core_fragments = set()
        for frag in theoretical_set:
            if not frag.is_extended:
                core_fragments.add(frag.name)

        total = 0
        core_matches = set()
        extended_matches = set()

        for peak_pair in self.solution_map:
            if peak_pair.fragment.series != series:
                continue
            fragment_name = peak_pair.fragment.base_name()
            if fragment_name in core_fragments:
                core_matches.add(fragment_name)
            else:
                extended_matches.add(fragment_name)
            peak = peak_pair.peak
            if peak.index.neutral_mass not in seen:
                seen.add(peak.index.neutral_mass)
                total += np.log10(peak.intensity) * (1 - (abs(peak_pair.mass_accuracy()) / error_tolerance) ** 4)
        glycan_composition = self.target.glycan_composition
        n = self._get_internal_size(glycan_composition)
        k = 2.0
        if not fragile_fucose:
            side_group_count = self._glycan_side_group_count(
                glycan_composition)
            if side_group_count > 0:
                k = 1.0
        d = max(n * np.log(n) / k, n)
        core_coverage = ((len(core_matches) * 1.0) / len(core_fragments)) ** glycan_core_weight
        extended_coverage = min(float(len(core_matches) + len(extended_matches)) / d, 1.0) ** glycan_coverage_weight
        score = total * core_coverage * extended_coverage
        self._glycan_coverage = core_coverage * extended_coverage
        glycan_prior = self.target.glycan_prior
        score += self._glycan_coverage * glycan_prior
        if np.isnan(score):
            return 0
        return score

    def peptide_score(self, error_tolerance=2e-5, peptide_coverage_weight=1.0, *args, **kwargs) -> float:
        if self._peptide_score is None:
            self._peptide_score = self.calculate_peptide_score(error_tolerance, peptide_coverage_weight, *args, **kwargs)
        return self._peptide_score

    def glycan_score(self, error_tolerance=2e-5, glycan_core_weight=0.4, glycan_coverage_weight=0.5,
                     *args, **kwargs) -> float:
        if self._glycan_score is None:
            self._glycan_score = self.calculate_glycan_score(
                error_tolerance, glycan_core_weight, glycan_coverage_weight, *args, **kwargs)
        return self._glycan_score


try:
    from glycresoft._c.tandem.tandem_scoring_helpers import calculate_peptide_score, calculate_glycan_score
    LogIntensityScorer.calculate_peptide_score = calculate_peptide_score
    LogIntensityScorer.calculate_glycan_score = calculate_glycan_score
except ImportError:
    pass


class LogIntensityScorerReweighted(LogIntensityScorer):
    def peptide_score(self, error_tolerance=2e-5, peptide_coverage_weight=0.7, *args, **kwargs):
        if self._peptide_score is None:
            self._peptide_score = self.calculate_peptide_score(
                error_tolerance, peptide_coverage_weight, *args, **kwargs)
        return self._peptide_score

    def peptide_coverage(self):
        return self._calculate_peptide_coverage_no_glycosylated()

try:
    from glycresoft._c.tandem.tandem_scoring_helpers import calculate_peptide_score_no_glycosylated
    LogIntensityScorerReweighted.calculate_peptide_score = calculate_peptide_score_no_glycosylated
except ImportError:
    pass


class ShortPeptideLogIntensityScorer(LogIntensityScorer):
    stub_weight = 0.65


def _short_peptide_test(scan, target, *args, **kwargs):
    return len(target) < 10


LogIntensityModelTree = ModelTreeNode(LogIntensityScorer, {
    _short_peptide_test: ModelTreeNode(ShortPeptideLogIntensityScorer, {}),
})


class FullSignaturePenalizedLogIntensityScorer(LogIntensityScorer):
    def calculate_glycan_score(self, error_tolerance=2e-5, core_weight=0.4, coverage_weight=0.5,
                               fragile_fucose=True, extended_glycan_search=False, *args, **kwargs):
        score = super(FullSignaturePenalizedLogIntensityScorer, self).calculate_glycan_score(
            error_tolerance, core_weight, coverage_weight, fragile_fucose=fragile_fucose,
            extended_glycan_search=extended_glycan_search, *args, **kwargs)
        signature_component = self._signature_ion_score()
        return score + signature_component


class HyperscoreScorer(SignatureAwareCoverageScorer, MassAccuracyMixin):

    def _calculate_hyperscore(self, *args, **kwargs):
        n_term_intensity = 0
        c_term_intensity = 0
        stub_intensity = 0
        n_term = 0
        c_term = 0
        stub_count = 0
        for peak, fragment in self.solution_map:
            if fragment.series == "oxonium_ion":
                continue
            elif fragment.series == IonSeries.stub_glycopeptide:
                stub_count += 1
                stub_intensity += peak.intensity
            elif fragment.series in (IonSeries.b, IonSeries.c):
                n_term += 1
                n_term_intensity += peak.intensity
            elif fragment.series in (IonSeries.y, IonSeries.z):
                c_term += 1
                c_term_intensity += peak.intensity
        hyper = 0
        factors = [math.factorial(n_term), math.factorial(c_term), math.factorial(stub_count),
                   stub_intensity, n_term_intensity, c_term_intensity]
        for f in factors:
            hyper += math.log(f)

        return hyper

    def calculate_score(self, error_tolerance=2e-5, *args, **kwargs):
        hyperscore = self._calculate_hyperscore(error_tolerance, *args, **kwargs)
        mass_accuracy = self._precursor_mass_accuracy_score()
        signature_component = self._signature_ion_score()
        self._score = hyperscore + mass_accuracy + signature_component

        return self._score



from collections import defaultdict
from glypy import Composition

mass_shift_index = dict()


class MassShiftBase(object):
    def __eq__(self, other):
        try:
            return (self.name == other.name and abs(
                self.mass - other.mass) < 1e-10) or self.composition == other.composition
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.name)

    def _register_name(self):
        mass_shift_index[self.name] = self.composition

    @classmethod
    def get(cls, name):
        return mass_shift_index[name]

try:
    from glycresoft._c.chromatogram_tree.mass_shift import MassShiftBase
    mass_shift_index = MassShiftBase._get_name_registry()
except ImportError:
    pass


class MassShift(MassShiftBase):
    def __init__(self, name, composition, tandem_composition=None, charge_carrier=0):
        self.name = name
        self.composition = composition
        self.mass = composition.mass
        self.charge_carrier = charge_carrier
        if tandem_composition is None:
            tandem_composition = self.composition.copy()
        self.tandem_composition = tandem_composition
        self.tandem_mass = self.tandem_composition.mass
        self._register_name()

    def __repr__(self):
        return "MassShift(%s, %s)" % (self.name, self.composition)

    def __mul__(self, n):
        if self.composition == {}:
            return self
        if isinstance(n, int):
            return CompoundMassShift({self: n})
        else:
            raise TypeError("Cannot multiply MassShift by non-integer")

    def __rmul__(self, n):
        if self.composition == {}:
            return self
        if isinstance(n, int):
            return CompoundMassShift({self: n})
        else:
            raise TypeError("Cannot multiply MassShift by non-integer")

    def __getstate__(self):
        return {
            "tandem_composition": self.tandem_composition,
            "name": self.name,
            "mass": self.mass,
            "charge_carrier": self.charge_carrier,
            "tandem_mass": self.tandem_mass,
            "composition": self.composition
        }

    def __setstate__(self, state):
        if isinstance(state, tuple):
            # We're receiving an older Cython pickled state
            self._hash = state[0]
            self.composition = state[1]
            self.mass = state[2]
            self.name = state[3]
            self.tandem_composition = state[4]
            self.tandem_mass = state[5]
            self.charge_carrier = state[6].get('charge_carrier', 0)

        else:
            self.name = state['name']
            self.charge_carrier = state.get('charge_carrier', 0)
            self.mass = state['mass']
            self.composition = state['composition']
            self.tandem_mass = state['tandem_mass']
            self.tandem_composition = state['tandem_composition']
            self._hash = hash(self.name)

    def __add__(self, other):
        if self.composition == {}:
            return other
        elif other.composition == {}:
            return self

        if self == other:
            return self * 2
        if isinstance(other, CompoundMassShift):
            return other + self
        return CompoundMassShift({
            self: 1,
            other: 1
        })

    def __sub__(self, other):
        if other.composition == {}:
            return self
        # if self.composition == {}:
        #     name = "-(%s)" % other.name
        #     composition = -other.composition
        #     tandem_composition = -other.tandem_composition
        #     charge_carrier = -other.charge_carrier
        if self == other:
            return Unmodified
        if isinstance(other, CompoundMassShift):
            return other - self
        else:
            return CompoundMassShift({
                self: 1,
                other: -1
            })

    def composed_with(self, other):
        return self == other


class CompoundMassShift(MassShiftBase):
    def __init__(self, counts=None):
        if counts is None:
            counts = {}
        self.counts = defaultdict(int, counts)
        self.composition = None
        self.tandem_composition = None
        self.name = None
        self.mass = 0
        self.tandem_mass = 0
        self.charge_carrier = 0
        self._compute_composition()
        self._compute_name()

    def __reduce__(self):
        return self.__class__, (self.counts, )

    def _compute_composition(self):
        composition = Composition()
        tandem_composition = Composition()
        charge_carrier = 0
        for k, v in self.counts.items():
            composition += k.composition * v
            tandem_composition += k.tandem_composition * v
            charge_carrier += k.charge_carrier * v
        self.composition = composition
        self.mass = composition.mass
        self.tandem_composition = tandem_composition
        self.tandem_mass = tandem_composition.mass
        self.charge_carrier = charge_carrier

    def _compute_name(self):
        parts = []
        for k, v in self.counts.items():
            if v == 1:
                parts.append(k.name)
            else:
                parts.append("%s * %d" % (k.name, v))
        self.name = " + ".join(sorted(parts))

    def composed_with(self, other):
        if isinstance(other, MassShift):
            return self.counts.get(other, 0) >= 1
        elif isinstance(other, CompoundMassShift):
            for key, count in other.counts.items():
                if self.counts.get(key, 0) != count:
                    return False
            return True

    def __add__(self, other):
        if other == Unmodified:
            return self
        elif self == Unmodified:
            return other

        if isinstance(other, MassShift):
            counts = defaultdict(int, self.counts)
            counts[other] += 1
            if counts[other] == 0:
                counts.pop(other)
            if counts:
                return self.__class__(counts)
            return Unmodified
        elif isinstance(other, CompoundMassShift):
            counts = defaultdict(int, self.counts)
            for k, v in other.counts.items():
                if v != 0:
                    counts[k] += v
                if counts[k] == 0:
                    counts.pop(k)
            if counts:
                return self.__class__(counts)
            return Unmodified
        else:
            return NotImplemented

    def __sub__(self, other):
        if other == Unmodified:
            return self
        if not self.composed_with(other):
            raise ValueError(
                "Cannot subtract %r from %r, not part of the compound" % (other, self))

        if isinstance(other, MassShift):
            counts = defaultdict(int, self.counts)
            counts[other] -= 1
            if counts[other] == 0:
                counts.pop(other)
            if counts:
                return self.__class__(counts)
            return Unmodified
        elif isinstance(other, CompoundMassShift):
            counts = defaultdict(int, self.counts)
            for k, v in other.counts.items():
                counts[k] -= v
                if counts[k] == 0:
                    counts.pop(k)
            if counts:
                return self.__class__(counts)
            return Unmodified
        else:
            return NotImplemented

    def __mul__(self, i):
        if self.composition == {}:
            return self
        if isinstance(i, int):
            counts = defaultdict(int, self.counts)
            for k in counts:
                if k == Unmodified:
                    continue
                counts[k] *= i
            return self.__class__(counts)
        else:
            raise TypeError("Cannot multiply MassShift by non-integer")

    def __neg__(self):
        return self * -1

    def __repr__(self):
        return "MassShift(%s, %s)" % (self.name, self.composition)


Unmodified = MassShift("Unmodified", Composition())
Formate = MassShift("Formate", Composition('HCOOH'), charge_carrier=1)
Ammonium = MassShift("Ammonium", Composition("NH3"), Composition())
Sodium = MassShift("Sodium", Composition("Na1H-1"), charge_carrier=1)
Potassium = MassShift("Potassium", Composition("K1H-1"), charge_carrier=1)
Deoxy = MassShift("Deoxy", Composition("O-1"), Composition())


class MassShiftCollection(object):
    def __init__(self, mass_shifts):
        self.mass_shifts = list(mass_shifts)
        self.mass_shift_map = {}
        self._invalidate()

    def _invalidate(self):
        self.mass_shift_map = {
            mass_shift.name: mass_shift for mass_shift in self.mass_shifts
        }

    def append(self, mass_shift):
        self.mass_shifts.append(mass_shift)
        self._invalidate()

    def __getitem__(self, i):
        try:
            return self.mass_shifts[i]
        except (IndexError, TypeError):
            return self.mass_shift_map[i]

    def __iter__(self):
        return iter(self.mass_shifts)

    def __len__(self):
        return len(self.mass_shifts)


import logging
try:
    import psycopg2
    DEC2FLOAT = psycopg2.extensions.new_type(
        psycopg2.extensions.DECIMAL.values,
        'DEC2FLOAT',
        lambda value, curs: float(value) if value is not None else None)
    psycopg2.extensions.register_type(DEC2FLOAT)
except ImportError:
    pass
logger = logging.getLogger("sqlalchemy.pool.NullPool")
logger.propagate = False
logger.addHandler(logging.NullHandler())

from dill._dill import _reverse_typemap
_reverse_typemap['IntType'] = int
_reverse_typemap['ListType'] = list


from glycresoft.task import TaskBase

from .match import (
    GlycanChromatogramMatcher,
    GlycopeptideChromatogramMatcher,
    NonSplittingChromatogramMatcher)

from .evaluate import (
    ChromatogramEvaluator,
    LogitSumChromatogramEvaluator,
    LaplacianRegularizedChromatogramEvaluator)


class ChromatogramProcessor(TaskBase):
    matcher_type = GlycanChromatogramMatcher

    def __init__(self, chromatograms, database, mass_shifts=None, mass_error_tolerance=1e-5,
                 scoring_model=None, smooth_overlap_rt=True, acceptance_threshold=0.4,
                 delta_rt=0.25, peak_loader=None):
        if mass_shifts is None:
            mass_shifts = []
        self._chromatograms = chromatograms
        self.database = database
        self.mass_shifts = mass_shifts
        self.mass_error_tolerance = mass_error_tolerance
        self.peak_loader = peak_loader

        self.scoring_model = scoring_model

        self.smooth_overlap_rt = smooth_overlap_rt
        self.acceptance_threshold = acceptance_threshold
        self.delta_rt = delta_rt

        self.solutions = None
        self.accepted_solutions = None

    def make_matcher(self):
        matcher = self.matcher_type(self.database)
        return matcher

    def _match_compositions(self):
        matcher = self.make_matcher()
        matches = matcher.process(
            self._chromatograms, self.mass_shifts, self.mass_error_tolerance,
            delta_rt=(self.delta_rt * 4 if self.smooth_overlap_rt else 0))
        return matches

    def make_evaluator(self):
        evaluator = ChromatogramEvaluator(self.scoring_model)
        return evaluator

    def match_compositions(self):
        self.log("Begin Matching Chromatograms")
        matches = self._match_compositions()
        self.log("End Matching Chromatograms")
        self.log("%d Chromatogram Candidates Found" % (len(matches),))
        return matches

    def evaluate_chromatograms(self, matches):
        self.log("Begin Evaluating Chromatograms")
        self.evaluator = self.make_evaluator()
        self.evaluator.configure({
            "peak_loader": self.peak_loader,
            "mass_shifts": self.mass_shifts,
            "delta_rt": self.delta_rt,
            "mass_error_tolerance": self.mass_error_tolerance,
            "matches": matches
        })
        self.solutions = self.evaluator.score(
            matches, smooth_overlap_rt=self.smooth_overlap_rt,
            mass_shifts=self.mass_shifts, delta_rt=self.delta_rt)
        self.accepted_solutions = self.evaluator.acceptance_filter(self.solutions)
        self.log("End Evaluating Chromatograms")
        return self.accepted_solutions

    def run(self):
        matches = self.match_compositions()
        result = self.evaluate_chromatograms(matches)
        return result

    def __iter__(self):
        if self.accepted_solutions is None:
            self.run()
        return iter(self.accepted_solutions)


class LogitSumChromatogramProcessor(ChromatogramProcessor):
    def make_evaluator(self):
        evaluator = LogitSumChromatogramEvaluator(self.scoring_model)
        return evaluator


class LaplacianRegularizedChromatogramProcessor(LogitSumChromatogramProcessor):
    GRID_SEARCH = 'grid'

    def __init__(self, chromatograms, database, network=None, mass_shifts=None, mass_error_tolerance=1e-5,
                 scoring_model=None, smooth_overlap_rt=True, acceptance_threshold=0.4,
                 delta_rt=0.25, peak_loader=None, smoothing_factor=0.2, grid_smoothing_max=1.0,
                 regularization_model=None):
        super(LaplacianRegularizedChromatogramProcessor, self).__init__(
            chromatograms, database, mass_shifts, mass_error_tolerance,
            scoring_model, smooth_overlap_rt, acceptance_threshold,
            delta_rt, peak_loader)
        if grid_smoothing_max is None:
            grid_smoothing_max = 1.0
        if network is None:
            network = database.glycan_composition_network
        self.network = network
        self.smoothing_factor = self.prepare_regularization_parameters(smoothing_factor)
        self.grid_smoothing_max = grid_smoothing_max
        self.regularization_model = regularization_model

    def prepare_regularization_parameters(self, smoothing_factor):
        if isinstance(smoothing_factor, tuple):
            if smoothing_factor[0] == self.GRID_SEARCH:
                smoothing_factor = (None, smoothing_factor[1])
        elif self.GRID_SEARCH == smoothing_factor:
            smoothing_factor = None
        return smoothing_factor

    def make_evaluator(self):
        evaluator = LaplacianRegularizedChromatogramEvaluator(
            self.scoring_model,
            self.network,
            smoothing_factor=self.smoothing_factor,
            grid_smoothing_max=self.grid_smoothing_max,
            regularization_model=self.regularization_model)
        return evaluator


class GlycopeptideChromatogramProcessor(ChromatogramProcessor):
    matcher_type = GlycopeptideChromatogramMatcher


class NonSplittingChromatogramProcessor(ChromatogramProcessor):
    matcher_type = NonSplittingChromatogramMatcher
